"""Collect of methods used for validating a BEL file."""
import csv
import difflib
import logging
import os
import re
from pathlib import Path
from textwrap import fill
from typing import Iterable, Optional, Union

import numpy as np
import pandas as pd

import ebel.database
from ebel.errors import BelSyntaxError
from ebel.parser import (bel_to_json, check_bel_script,
                         check_bel_script_line_by_line)

logger = logging.getLogger(__name__)


def validate_bel_file(
    bel_script_path: Union[str, Path],
    force_new_db: bool = False,
    line_by_line: bool = False,
    reports: Union[Iterable[str], str] = None,
    bel_version: str = "2_1",
    tree: bool = False,
    sqlalchemy_connection_str: str = None,
    json_file: bool = True,
    force_json: bool = False,
):
    """Validate BEL script for correct syntax following eBNF grammar.

    Parameters
    ----------
    bel_script_path: str
        Path to BEL file or directory contaiing BEL files.
    force_new_db: bool
        Delete current database of namespaces/values and generate a new one. Defaults to False.
    line_by_line: bool
        TODO: Write this.
    reports: Iterable[str] or str
        List of file paths to write reports to. Multiple formats of the report can be generated at once. Acceptable
        formats include: CSV, TSV, TXT, XLS, XLSX, JSON, HTML, MD
    bel_version: {'1', '2', '2_1'}
        Which BEL grammar version should be used for validating the BEL file. Current available are 1.0, 2.0, and 2.1.
        Defaults to the most recent version.
    tree: bool
        Generates a tree of relationships derived from the BEL file. Defaults to False.
    sqlalchemy_connection_str: str
        Path to SQLLite database to be used for storing/looking up used namespaces and values. If None given, it uses
        the generated e(BE:L) database (default).
    json_file: bool
        If True, generates a JSON file that can be used for importing BEL relationships into an e(BE:L) generated
        OrientDB database. Only creates the JSON file when there are no grammar or syntax errors. Defaults to True.
    force_json: bool
        If True, will create an importable JSON file even if there are namespace/value errors. Defaults to False.

    Returns
    -------
    dict
        Dictionary of file paths and results for each BEL file processed.

    Examples
    --------
    Task: Validate BEL script `my.bel` for BEL syntax 2.0, create error
    reports in Markdown and JSON format. In case of no errors create a JSON file
    for the import of BEL network into Cytoscape:

    > ebel validate my.bel -v 2 -r error_report.md,error_report.json

    """
    validation_results = dict()

    if line_by_line:
        # TODO: This is perhaps not working
        result = check_bel_script_line_by_line(
            bel_script_path, error_report_file_path=reports, bel_version=bel_version
        )

        if reports:
            logger.info("Wrote report to %s\n" % reports)
        else:
            logger.info("\n".join([x.to_string() for x in result]) + "\n")

    else:
        if sqlalchemy_connection_str:
            ebel.database.set_connection(sqlalchemy_connection_str)

        bel_files = _create_list_bel_files(bel_path=bel_script_path)
        validation_results["bel_files_checked"] = bel_files

        for bel_file in bel_files:
            # Create dict to be filled for individual BEL files.
            validation_results[bel_file] = dict()

            logger.info(f"Processing {bel_file}")
            result = check_bel_script(
                bel_script_path=bel_file,
                force_new_db=force_new_db,
                bel_version=bel_version,
            )

            if json_file:
                if result["errors"]:
                    if force_json:  # Check for syntax errors
                        bel_syntax_error_present = result["errors"] and any(
                            [isinstance(error_type, BelSyntaxError) for error_type in result["errors"]]
                        )
                        if bel_syntax_error_present:
                            logger.error("Cannot force JSON file due to syntax errors. Please check the BEL file.")

                        else:
                            json_file = _write_odb_json(
                                bel_path=bel_file,
                                results=result,
                                bel_version=bel_version,
                            )
                            validation_results[bel_file]["json"] = json_file

                    else:
                        logger.error("Unable to create JSON file due to grammar/syntax errors in BEL file")

                else:  # No errors so everything is fine
                    json_file = _write_odb_json(bel_path=bel_file, results=result, bel_version=bel_version)
                    validation_results[bel_file]["json"] = json_file

            if tree:
                if result["errors"]:
                    logger.error("Tree can not be printed because errors still exists\n")

                else:
                    logger.debug(result["tree"])
                    validation_results[bel_file]["tree"] = result["tree"]

                if result["warnings"] and reports:
                    report_paths = _write_report(reports, result, report_type="warnings")
                    validation_results[bel_file]["reports"] = report_paths

            elif result["errors"]:
                if not reports:
                    logger.info("\n".join([x.to_string() for x in result["errors"]]) + "\n")

                else:
                    _write_report(reports, result, report_type="errors")


def repair_bel_file(bel_script_path: str, new_file_path: Optional[str] = None):
    """Repair a BEL document.

    Parameters
    ----------
    bel_script_path : str
        Path to the BEL file.
    new_file_path : str (optional)
        Export repaired version of file to new path.
    """
    # if evidence:
    # regular expression for missing continuous line (\ at the end of line)
    with open(bel_script_path, "r", encoding="utf-8") as belfile:
        content = belfile.read()

    new_content = content

    for regex_pattern in re.findall(
        r"\n((SET\s+(DOCUMENT\s+Description|Evidence|SupportingText|Support)" r'\s*=\s*)"(((?<=\\)"|[^"])+)"\s*\n*)',
        content,
    ):
        if regex_pattern[2].startswith("DOCUMENT"):
            new_prefix = "SET DOCUMENT Description = "
        else:
            new_prefix = "SET Support = "

        new_evidence_text = re.sub(r"(\\?[\r\n]+)|\\ ", " ", regex_pattern[3].strip())
        new_evidence_text = re.sub(r"\s{2,}", " ", new_evidence_text)
        new_evidence_text = re.sub(r"(\\)(\w)", r"\g<2>", new_evidence_text)
        new_evidence_text = fill(new_evidence_text, break_long_words=False).replace("\n", " \\\n")
        new_evidence = new_prefix + '"' + new_evidence_text + '"\n\n'

        new_content = new_content.replace(regex_pattern[0], new_evidence)

    if content != new_content:
        if new_file_path:
            with open(new_file_path + ".diff2repaired", "w") as new_file:
                new_file.write("\n".join(list(difflib.ndiff(content.split("\n"), new_content.split("\n")))))

        else:
            with open(bel_script_path, "w") as output_file:
                output_file.write(new_content)


def _write_odb_json(bel_path: str, results: dict, bel_version: str) -> str:
    json_path = bel_path + ".json"
    if int(bel_version[0]) > 1:
        json_tree = bel_to_json(results["tree"])
        open(json_path, "w").write(json_tree)

    return json_path


def _create_list_bel_files(bel_path: str) -> list:
    """Export all BEL files in directory as list. If single file is passed, returns a list with that path."""
    if os.path.isdir(bel_path):
        bel_files = []
        for file in os.listdir(bel_path):
            if file.endswith(".bel"):
                bel_file_path = os.path.join(bel_path, file)
                bel_files.append(bel_file_path)

    else:
        bel_files = [bel_path]

    return bel_files


def _write_report(reports: Union[Iterable[str], str], result: dict, report_type: str) -> list:
    """Write report in different types depending on the file name suffix in reports.

    Parameters
    ----------
    reports : Iterable[str] or str
        List of report formats or comma separated list of report file names.
    result : dict
        return value of check_bel_script methode.
    report_type : str
        `report_type` could be 'warnings' or 'errors'.

    Returns
    -------
    list
        List of file paths for the reports written.

    """
    # TODO: report_type options should be constants
    errors_or_warns_as_list_of_dicts = [x.to_dict() for x in result[report_type]]

    columns = [
        report_type[:-1] + "_class",
        "url",
        "keyword",
        "entry",
        "line_number",
        "column",
        "hint",
    ]
    df = pd.DataFrame(data=errors_or_warns_as_list_of_dicts, columns=columns)
    df.index += 1

    if isinstance(reports, str):
        reports = reports.split(",")

    for report in reports:
        try:
            if report.endswith(".csv"):
                df.to_csv(report, index=False)

            if report.endswith(".xls"):
                try:
                    df.to_excel(report, index=False)

                except ValueError:
                    logger.warning("Max Excel sheet size exceeded. Writing to CSV instead.")
                    df.to_csv(report, index=False)

            if report.endswith(".xlsx"):
                try:
                    df.to_excel(report, engine="xlsxwriter", index=False)

                except ValueError:
                    logger.warning("Max Excel sheet size exceeded. Writing to CSV instead.")
                    df.to_csv(report, index=False)

            if report.endswith(".tsv"):
                df.to_csv(report, sep="\t", index=False)

            if report.endswith(".json"):
                df.to_json(report, index=False)

            if report.endswith(".txt"):
                open(report, "w").write(df.to_string(index=False))

            if report.endswith(".html"):
                df.to_html(report, index=False)

            if report.endswith(".md"):
                cols = df.columns
                df2 = pd.DataFrame(
                    [
                        [
                            "---",
                        ]
                        * len(cols)
                    ],
                    columns=cols,
                )

                if df.hint.dtype == np.str:
                    df.hint = df.hint.str.replace(r"\|", "&#124;")

                if df.entry.dtype == np.str:
                    df.entry = df.entry.str.replace(r"\|", "&#124;")

                df.url = [("[url](" + str(x) + ")" if not pd.isna(x) else "") for x in df.url]
                url_template = "[%s](" + report.split(".bel.")[0] + ".bel?expanded=true&viewer=simple#L%s)"
                df.line_number = [url_template % (x, x) for x in df.line_number]
                df3 = pd.concat([df2, df])
                df3.to_csv(
                    report,
                    sep="|",
                    index=False,
                    quoting=csv.QUOTE_NONE,
                    escapechar="\\",
                )

        except PermissionError:
            logger.error("Previous version of error report is still open and cannot be overwritten. Unable to update.")

    return reports



"""True values to check against."""

# True values after importing basic_import_test.bel.json with only extension - test_import_json
NODES_EXTENSION = {
    "bel": 12,
    "nn": 12,
    "pure_object": 11,
    "location_object": 10,
    "bio_concept": 0,
    "biological_process": 0,
    "pathology": 0,
    "bio_object": 11,
    "complex": 0,
    "abundance": 1,
    "population": 0,
    "genetic_flow": 10,
    "gene": 4,
    "rna": 3,
    "protein": 3,
    "micro_rna": 0,
    "bio_act": 1,
    "activity": 1,
    "reaction": 0,
    "degradation": 0,
    "cell_secretion": 0,
    "translocation": 0,
    "cell_surface_expression": 0,
    "bio_list": 0,
    "list": 0,
    "composite": 0,
    "ebel": 0,
    "variant": 0,
    "fragment": 0,
    "location": 0,
    "pmod": 0,
    "gmod": 0,
    "from_location": 0,
    "to_location": 0,
    "reactants": 0,
    "products": 0,
    "fusion_protein": 0,
    "fusion_rna": 0,
    "fusion_gene": 0,
}
EDGES_EXTENSION = {
    "bel_relation": 9,
    "causal": 3,
    "increases": 1,
    "directly_increases": 1,
    "decreases": 1,
    "directly_decreases": 0,
    "rate_limiting_step_of": 0,
    "causes_no_change": 0,
    "regulates": 0,
    "correlative": 0,
    "negative_correlation": 0,
    "positive_correlation": 0,
    "association": 0,
    "no_correlation": 0,
    "genomic": 6,
    "orthologous": 0,
    "transcribed_to": 3,
    "translated_to": 3,
    "other": 0,
    "has_member": 0,
    "has_members": 0,
    "has_component": 0,
    "has_components": 0,
    "equivalent_to": 0,
    "is_a": 0,
    "sub_process_of": 0,
    "deprecated": 0,
    "analogous_to": 0,
    "biomarker_for": 0,
    "prognostic_biomarker_for": 0,
    "compiler": 0,
    "acts_in": 0,
    "has_product": 0,
    "has_variant": 0,
    "has_modification": 0,
    "reactant_in": 0,
    "translocates": 0,
    "includes": 0,
    "ebel_relation": 1,
    "has__protein": 1,
    "has__rna": 0,
    "has__gene": 0,
    "has__abundance": 0,
    "has__population": 0,
    "has__location": 0,
    "has__from_location": 0,
    "has__to_location": 0,
    "has__fragment": 0,
    "has__pmod": 0,
    "has__gmod": 0,
    "has__complex": 0,
    "has__micro_rna": 0,
    "has__variant": 0,
    "has__reactants": 0,
    "has__products": 0,
    "has__composite": 0,
    "has_fragmented_protein": 0,
    "has_modified": 0,
    "has_modified_protein": 0,
    "has_modified_gene": 0,
    "has_variant_obj": 0,
    "has_variant_gene": 0,
    "has_variant_rna": 0,
    "has_variant_protein": 0,
    "has_variant_micro_rna": 0,
    "has_located": 0,
    "has_located_gene": 0,
    "has_located_rna": 0,
    "has_located_protein": 0,
    "has_located_micro_rna": 0,
    "has_located_complex": 0,
    "has_located_abundance": 0,
    "has_located_population": 0,
    "pathway_interaction": 0,
    "has_ppi": 0,
}



"""Expression Atlas API methods."""
import json
import math
import re
from collections import Counter

from flask import request
from sqlalchemy import inspect

from ebel import Bel
from ebel.manager.rdbms.models.expression_atlas import (Experiment, FoldChange,
                                                        GroupComparison, Gsea,
                                                        Idf, SdrfCondensed)
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import _get_data

from . import _get_paginated_query_result, add_query_filters

models_dict = {
    "experiment": Experiment,
    "idf": Idf,
    "gsea": Gsea,
    "group_comparison": GroupComparison,
    "foldchange": FoldChange,
    "sdrf_condensed": SdrfCondensed,
}


def get_expriment():
    """Get list of experiments."""
    return _get_data(Experiment)


def get_idf():
    """Get list of IDFs."""
    return _get_data(Idf)


def get_group_comparison():
    """Get list of group comparisons."""
    return _get_data(GroupComparison)


def get_fold_change():
    """Get list of fold changes."""
    return _get_data(FoldChange)


def get_sdrf_condensed():
    """Get list of condensed version of SDRFs."""
    return _get_data(SdrfCondensed)


def get_gsea():
    """Get GSEA data."""
    return _get_data(Gsea, order_by=[Gsea.p_adj_non_dir])


class GseaType:
    """String constant definitions."""

    GO = "go"
    REACTOME = "reactome"
    INTERPRO = "interpro"


def get_most_common_gseas_by_group_comparison_ids() -> list:
    """Identify most often ocurring GSEAs based on group comparison IDs."""
    b = Bel()
    group_comparison_ids_str = request.args.get("group_comparison_ids", "")
    group_comparison_ids = [
        int(x.strip())
        for x in group_comparison_ids_str.split(",")
        if isinstance(x, str) and re.search(r"\s*\d+\s*", x)
    ]
    query = b.session.query(Gsea.accession, Gsea.term, Gsea.gsea_type).filter(
        Gsea.group_comparison_id.in_(group_comparison_ids)
    )

    gsea_type = request.args.get("gsea_type", "")
    gsea_type = gsea_type if gsea_type in [GseaType.GO, GseaType.REACTOME, GseaType.INTERPRO] else ""

    if gsea_type:
        query = query.filter(Gsea.gsea_type == gsea_type)

    p_adj_non_dir = request.args.get("p_adj_non_dir", "")
    p_adj_non_dir = float(p_adj_non_dir) if re.search(r"\s*\d+(\.\d+)?\s*", p_adj_non_dir) else None
    if isinstance(p_adj_non_dir, float):
        query = query.filter(Gsea.p_adj_non_dir <= p_adj_non_dir)

    results = [x for x in query.all()]

    return [
        {"accession": x[0][0], "term": x[0][1], "type": x[0][2], "occurence": x[1]}
        for x in Counter(results).most_common()
        if x[1] > 1
    ]


def get_comparison_groups_by_genes():
    """Get experiments, comparion groups and foldchanges by 2 genes."""
    gene_1 = request.args.get("gene_symbol_1")
    gene_2 = request.args.get("gene_symbol_2")
    return _get_comparison_groups_by_genes(gene_1, gene_2)


def _get_comparison_groups_by_genes(gene_1, gene_2):
    """Get experiments, comparion groups and foldchanges by 2 genes."""
    result_keys = [
        "fold_change_gene_symbol_1",
        "fold_change_gene_symbol_2",
        "group_comparison",
        "experiment_title",
        "experiment_identifier",
        "is_positive_correlated",
    ]
    sql = f"""Select
        a.log2foldchange,
        b.log2foldchange,
        gc.name,
        e.title,
        e.name,
        if(((a.log2foldchange>0 and b.log2foldchange>0) or (a.log2foldchange<0 and b.log2foldchange<0)),1,0)
    from (Select
            group_comparison_id,
            log2foldchange
        from
            expression_atlas_foldchange where gene_name = '{gene_1}') a
    inner join (Select
            group_comparison_id,
            log2foldchange
        from expression_atlas_foldchange where gene_name = '{gene_2}') b
    using(group_comparison_id)
    inner join expression_atlas_group_comparison gc on (gc.id=group_comparison_id)
    inner join expression_atlas_experiment e on (e.id = gc.experiment_id)
    order by (a.log2foldchange + b.log2foldchange) desc"""
    with Bel().engine.connect() as con:
        return list([dict(zip(result_keys, x)) for x in con.execute(sql).fetchall()])


def get_comparison_groups_by_edge_rid():
    """Use edge rIDs to get different groups to compare."""
    rid = request.args.get("edge_rid")
    res = Bel().query_get_dict(
        f"""Select
                                    in.name as name_in,
                                    in.namespace as ns_in,
                                    out.name as name_out,
                                    out.namespace as ns_out,
                                    in.bel as in_bel,
                                    out.bel as out_bel,
                                    @class.asString() as relation,
                                    citation,
                                    annotation,
                                    evidence
                                from {rid}"""
    )
    if res:
        has_all_cols = all([(x in res[0]) for x in ["name_in", "ns_in", "name_out", "ns_out"]])
        if has_all_cols:
            both_ns_hgnc = res[0]["ns_in"] == "HGNC" and res[0]["ns_out"] == "HGNC"
            not_the_same = res[0]["name_in"] != res[0]["name_out"]
            if has_all_cols and both_ns_hgnc and not_the_same:
                gene_symbol_1 = res[0]["name_in"]
                gene_symbol_2 = res[0]["name_out"]
                comparisons = _get_comparison_groups_by_genes(gene_symbol_1, gene_symbol_2)
                return {
                    "gene_symbol_1": gene_symbol_1,
                    "gene_symbol_2": gene_symbol_2,
                    "citation": res[0].get("citation"),
                    "annotation": res[0].get("annotation"),
                    "evidence": res[0].get("evidence"),
                    "relation": res[0].get("relation"),
                    "comparisons": comparisons,
                    "bel_object": res[0].get("in_bel"),
                    "bel_subject": res[0].get("out_bel"),
                }
    return []


def get_expression_atlas():
    """Get EA results."""
    columns = [
        Experiment.name,
        GroupComparison.name,
        Experiment.title,
        GroupComparison.group_comparison,
        GroupComparison.id,
    ]
    data = json.loads(request.data)
    b = Bel()
    query = b.session.query(Experiment).join(GroupComparison)
    for table, columns_params in [(k, v) for k, v in data.items() if k not in ("page", "page_size")]:
        any_value = any([v["value"].strip() for k, v in columns_params.items()])
        if any_value:
            model = models_dict[table]

            if model not in (Experiment, GroupComparison):
                query = query.join(model)
            query = add_query_filters(query, columns_params, model)
    query = query.with_entities(*columns)
    query = query.group_by(Experiment.id, GroupComparison.id)
    limit = int(data["page_size"]) if data.get("page_size") else 10
    page = int(data["page"]) if data.get("page") else 1
    count = query.count()
    query = query.limit(limit).offset(limit * (page - 1))

    print(query.statement.compile(compile_kwargs={"literal_binds": True}))
    column_names = [f"{x.parent.class_.__tablename__}.{x.name}" for x in [inspect(c) for c in columns]]
    return {
        "page": page,
        "pages": math.ceil(count / limit),
        "count": count,
        "page_size": limit,
        "column_names": column_names,
        "results": [tuple(x) for x in query.all()],
    }



"""StringDB API methods."""

from flask import request

from ebel.manager.rdbms.models.stringdb import StringDb, StringDbAction
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result


def get_stringdb():
    """Get generic StringDB entry."""
    return _get_data(StringDb)


def get_stringdb_action():
    """Get StringDB entry by action."""
    return _get_data(StringDbAction)


def get_ebel_relation():
    """Get StringDB related eBEL relations."""
    conf = {
        "interactor_a_rid": "out.@rid.asString()",
        "interactor_a_name": "out.name",
        "interactor_a_namespace": "out.namespace",
        "interactor_a_uniprot": "out.uniprot",
        "edge_id": "@rid.asString()",
        "relation": "@class.asString()",
        "interactor_b_rid": "in.@rid.asString()",
        "interactor_b_namespace": "in.namespace",
        "interactor_b_name": "in.name",
        "interactor_b_uniprot": "in.uniprot",
    }
    sql = "SELECT "
    sql += ", ".join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_action_st"

    ra = request.args
    paras = {k: ra[k] for k in ra if k in conf}

    if paras:
        wheres = [f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()]
        sql += " WHERE " + " AND ".join(wheres)

    return _get_paginated_ebel_query_result(sql)



"""NCBI API methods."""

from flask import request

from ebel.manager.rdbms.models import ncbi
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_query_result


def get_gene_by_go():
    """Get gene information using GO."""
    return _get_data(ncbi.NcbiGeneGo)


def get_gene_by_mim():
    """Get gene information using MIM."""
    return _get_data(ncbi.NcbiGeneMim)


def get_gene_by_ensembl():
    """Get gene information using EnsEMBL."""
    return _get_data(ncbi.NcbiGeneEnsembl)


def get_gene_info():
    """Get gene information."""
    return _get_data(ncbi.NcbiGeneInfo)


def get_medgen():
    """Get MedGen terms."""
    return _get_data(ncbi.NcbiMedGenName)


def get_medgen_by_pmid():
    """Get MedGen terms by PMID."""
    n = ncbi.NcbiMedGenName
    p = ncbi.NcbiMedGenPmid
    q = RDBMS.get_session().query(n.cui, n.name).join(p).filter_by(pmid=request.args.get("pmid"))
    return _get_paginated_query_result(q, return_dict=True)


def get_go_by_pmid():
    """Get gene ontology by PMID."""
    q = RDBMS.get_session().query(ncbi.NcbiGeneGo).join(ncbi.NcbiGeneGoPmid).filter_by(pmid=request.args.get("pmid"))
    return _get_paginated_query_result(q, print_sql=True)



"""Reactome API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestReactome:
    def test_get_reactome(self, client):
        response = client.get(
            "api/v1/reactome?identifier=R-HSA-198933&uniprot_accession=P20138&organism=Homo%20sapiens&name=Immunoregulatory%20interactions%20between%20a%20Lymphoid%20and%20a%20non-Lymphoid%20cell&evidence_type=TAS&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "evidence_type": "TAS",
            "id": 140784,
            "identifier": "R-HSA-198933",
            "name": "Immunoregulatory interactions between a Lymphoid and a non-Lymphoid cell",
            "organism": "Homo sapiens",
            "uniprot_accession": "P20138",
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""PSI-MI 25 Molecular interaction types (https://www.ebi.ac.uk/ols/ontologies/mi)."""

types = {
    0: "molecular interaction",
    1: "interaction detection method",
    2: "participant identification method",
    3: "feature detection method",
    4: "affinity chromatography technology",
    5: "alanine scanning",
    6: "anti bait coimmunoprecipitation",
    7: "anti tag coimmunoprecipitation",
    8: "array technology",
    9: "bacterial display",
    10: "beta galactosidase complementation",
    11: "beta lactamase complementation",
    12: "bioluminescence resonance energy transfer",
    13: "biophysical",
    14: "adenylate cyclase complementation",
    16: "circular dichroism",
    17: "classical fluorescence spectroscopy",
    18: "two hybrid",
    19: "coimmunoprecipitation",
    20: "transmission electron microscopy",
    21: "colocalization by fluorescent probes cloning",
    22: "colocalization by immunostaining",
    23: "colocalization/visualisation technologies",
    24: "confirmational text mining",
    25: "copurification",
    26: "correlated mutations",
    27: "cosedimentation",
    28: "cosedimentation in solution",
    29: "cosedimentation through density gradient",
    30: "cross-linking study",
    31: "protein cross-linking with a bifunctional reagent",
    32: "de novo protein sequencing by mass spectrometry",
    33: "deletion analysis",
    34: "display technology",
    35: "docking",
    36: "domain fusion",
    37: "domain profile pairs",
    38: "dynamic light scattering",
    39: "edman degradation",
    40: "electron microscopy",
    41: "electron nuclear double resonance",
    42: "electron paramagnetic resonance",
    43: "electron resonance",
    45: "experimental interaction detection",
    46: "experimental knowledge based",
    47: "far western blotting",
    48: "filamentous phage display",
    49: "filter binding",
    50: "flag tag coimmunoprecipitation",
    51: "fluorescence technology",
    52: "fluorescence correlation spectroscopy",
    53: "fluorescence polarization spectroscopy",
    54: "fluorescence-activated cell sorting",
    55: "fluorescent resonance energy transfer",
    56: "full identification by DNA sequencing",
    57: "gene neighbourhood",
    58: "genome based prediction",
    59: "gst pull down",
    60: "ha tag coimmunoprecipitation",
    61: "his pull down",
    62: "his tag coimmunoprecipitation",
    63: "interaction prediction",
    64: "interologs mapping",
    65: "isothermal titration calorimetry",
    66: "lambda phage display",
    67: "light scattering",
    68: "mass detection of residue modification",
    69: "mass spectrometry studies of complexes",
    70: "mobility shift",
    71: "molecular sieving",
    72: "monoclonal antibody western blot",
    73: "mrna display",
    74: "mutation analysis",
    75: "myc tag coimmunoprecipitation",
    76: "neural network on interface properties",
    77: "nuclear magnetic resonance",
    78: "nucleotide sequence identification",
    79: "other biochemical technologies",
    80: "partial DNA sequence identification by hybridization",
    81: "peptide array",
    82: "peptide massfingerprinting",
    83: "peptide synthesis",
    84: "phage display",
    85: "phylogenetic profile",
    86: "polyclonal antibody western blot",
    87: "predictive text mining",
    88: "primer specific pcr",
    89: "protein array",
    90: "protein complementation assay",
    91: "chromatography technology",
    92: "protein in situ array",
    93: "protein sequence identification",
    94: "protein staining",
    95: "proteinchip(r) on a surface-enhanced laser desorption/ionization",
    96: "pull down",
    97: "reverse ras recruitment system",
    98: "ribosome display",
    99: "scintillation proximity assay",
    100: "sequence based phylogenetic profile",
    101: "sequence based prediction",
    102: "sequence tag identification",
    103: "southern blot",
    104: "static light scattering",
    105: "structure based prediction",
    106: "surface patches",
    107: "surface plasmon resonance",
    108: "t7 phage display",
    109: "tap tag coimmunoprecipitation",
    110: "text mining",
    111: "dihydrofolate reductase reconstruction",
    112: "ubiquitin reconstruction",
    113: "western blot",
    114: "x-ray crystallography",
    115: "yeast display",
    116: "feature type",
    117: "binding-associated region",
    118: "mutation",
    119: "mutation decreasing interaction",
    120: "post translation modification",
    121: "acetylated residue",
    122: "n-acetyl-alanine",
    123: "n2-acetyl-arginine",
    124: "n-acetyl-asparagine",
    125: "n-acetyl-aspartic acid",
    126: "n-acetyl-cysteine",
    127: "n-acetyl-glutamine",
    128: "n-acetyl-glutamic acid",
    129: "n-acetylglycine",
    130: "n-acetyl-histidine",
    131: "n-acetyl-isoleucine",
    132: "n-acetyl-leucine",
    133: "n2-acetyl-lysine",
    134: "n6-acetyl-lysine",
    135: "n-acetyl-methionine",
    136: "n-acetyl-phenylalanine",
    137: "n-acetyl-proline",
    138: "n-acetyl-serine",
    139: "n-acetyl-threonine",
    140: "n-acetyl-tryptophan",
    141: "n-acetyl-tyrosine",
    142: "n-acetyl-valine",
    143: "amidated residue",
    145: "arginine amide",
    146: "formylated residue",
    147: "n-formyl-methionine",
    148: "hydroxylated residue",
    150: "lipid modification",
    151: "s-farnesyl-cysteine",
    152: "s-geranylgeranyl-cysteine",
    153: "n-palmitoyl-cysteine",
    154: "s-palmitoyl-cysteine",
    155: "n-myristoyl-glycine",
    156: "n6-myristoyl-lysine",
    157: "methylated residue",
    158: "n-methyl-alanine",
    159: "n,n,n-trimethyl-alanine",
    160: "omega-n,omega-n-dimethyl-arginine",
    161: "beta-methylthioaspartic acid",
    162: "n5-methyl-glutamine",
    163: "glutamic acid 5-methyl ester",
    165: "n6-methyl-lysine",
    166: "n6,n6-dimethyl-lysine",
    167: "n6,n6,n6-trimethyl-lysine",
    168: "n-methyl-methionine",
    169: "n-methyl-phenylalanine",
    170: "phosphorylated residue",
    171: "omega-n-phospho-arginine",
    172: "aspartic 4-phosphoric anhydride",
    173: "s-phospho-cysteine",
    176: "o-phospho-serine",
    177: "o-phospho-threonine",
    178: "o4'-phospho-tyrosine",
    179: "other modification",
    180: "selenocysteine",
    181: "selenomethionine",
    182: "3-oxoalanine",
    184: "glutamyl 5-glycerylphosphorylethanolamine",
    186: "n6-biotinyl-lysine",
    187: "n6-(4-amino-2-hydroxybutyl)-lysine",
    188: "n6-retinal-lysine",
    189: "ubiquitinated lysine",
    190: "interaction type",
    191: "aggregation",
    192: "acetylation reaction",
    193: "amidation reaction",
    194: "cleavage reaction",
    195: "covalent binding",
    196: "covalent interaction",
    197: "deacetylation reaction",
    198: "defarnesylation reaction",
    199: "deformylation reaction",
    200: "degeranylation reaction",
    201: "demyristoylation reaction",
    202: "depalmitoylation reaction",
    203: "dephosphorylation reaction",
    204: "deubiquitination reaction",
    205: "disaggregation",
    206: "farnesylation reaction",
    207: "formylation reaction",
    208: "genetic interaction",
    209: "geranylgeranylation reaction",
    210: "hydroxylation reaction",
    211: "lipid addition",
    212: "lipoprotein cleavage reaction",
    213: "methylation reaction",
    214: "myristoylation reaction",
    215: "non covalent interaction",
    216: "palmitoylation reaction",
    217: "phosphorylation reaction",
    218: "physical interaction",
    219: "synthetic lethal",
    220: "ubiquitination reaction",
    221: "expression level",
    222: "physiological level",
    223: "under expressed level",
    225: "chromatin immunoprecipitation array",
    226: "ion exchange chromatography",
    227: "reverse phase chromatography",
    228: "cytoplasmic complementation assay",
    230: "membrane bound complementation assay",
    231: "mammalian protein protein interaction trap",
    232: "transcriptional complementation assay",
    233: "protein dna complex",
    234: "131i radiolabel",
    235: "14c radiolabel",
    236: "32p radiolabel",
    237: "33p radiolabel",
    238: "3h radiolabel",
    239: "biotin tag",
    240: "fusion protein",
    241: "horseradish peroxidase tag",
    242: "gene ontology definition reference",
    243: "isoform parent sequence reference",
    244: "reactome complex",
    245: "reactome protein",
    246: "cabri",
    247: "newt",
    248: "resid",
    249: "huge",
    250: "gene",
    251: "gene product",
    252: "biological feature",
    253: "isotope label",
    254: "genetic interference",
    255: "post transcriptional interference",
    256: "rna interference",
    257: "antisense rna",
    258: "inhibitor antibodies",
    259: "perturbagens peptides",
    260: "inhibitor small molecules",
    261: "suppression",
    262: "suppression mutation",
    263: "suppression knockout",
    264: "suppression partial alteration",
    265: "suppression expression alteration",
    266: "suppression overexpression",
    267: "suppression scalable",
    268: "suppression underexpression",
    269: "synthetic phenotype",
    270: "conditional synthetic lethal",
    271: "conditional synthetic lethal temperature-sensitivity",
    273: "synthetic growth effect",
    274: "synthetic growth defect",
    275: "synthetic growth increase",
    276: "blue native page",
    300: "alias type",
    301: "gene name",
    302: "gene name synonym",
    303: "gene ontology synonym",
    304: "isoform synonym",
    305: "ordered locus name",
    306: "open reading frame name",
    307: "delivery method",
    308: "electroporation",
    309: "genomic tagging",
    310: "infection",
    311: "microinjection",
    312: "nucleic acid transfection",
    313: "interactor type",
    314: "complex",
    315: "protein complex",
    316: "ribonucleoprotein complex",
    317: "interaction",
    318: "nucleic acid",
    319: "deoxyribonucleic acid",
    320: "ribonucleic acid",
    321: "catalytic rna",
    322: "guide rna",
    323: "heterogeneous nuclear rna",
    324: "messenger rna",
    325: "transfer rna",
    326: "protein",
    327: "peptide",
    328: "small molecule",
    329: "unknown participant",
    330: "molecular source",
    331: "engineered",
    332: "naturally occurring",
    333: "feature range status",
    334: "c-terminal position",
    335: "certain sequence position",
    336: "greater-than",
    337: "less-than",
    338: "range",
    339: "undetermined sequence position",
    340: "n-terminal position",
    341: "ragged n-terminus",
    342: "sample process",
    343: "cdna library",
    344: "cell lysate",
    345: "author assigned name",
    346: "experimental preparation",
    348: "fixed cell",
    349: "living cell",
    350: "purified",
    351: "homogeneous",
    352: "partially purified",
    353: "cross-reference type",
    354: "cellular component",
    355: "molecular function",
    356: "identical object in an external resource",
    357: "method reference",
    358: "primary-reference",
    359: "biological process",
    360: "secondary accession number",
    361: "additional information",
    362: "inference",
    363: "inferred by author",
    364: "inferred by curator",
    365: "enzyme tag",
    366: "alkaline phosphatase tag",
    367: "green fluorescent protein tag",
    368: "yellow fluorescent protein tag",
    369: "lex-a dimerization assay",
    370: "tox-r dimerization assay",
    371: "35s radiolabel",
    372: "subcellular preparation",
    373: "dye label",
    374: "cyanine label",
    375: "cy3 label",
    376: "cy5 label",
    377: "fluorescein isothiocyanate label",
    378: "rare isotope label",
    379: "13c label",
    380: "15n label",
    381: "2h label",
    382: "mutation increasing interaction",
    383: "biopolymer",
    384: "alexa label",
    385: "alexa 350 label",
    386: "alexa 430 label",
    387: "alexa 488 label",
    388: "alexa 532 label",
    389: "alexa 546 label",
    390: "alexa 568 label",
    391: "alexa 594 label",
    396: "predetermined participant",
    397: "two hybrid array",
    398: "two hybrid pooling approach",
    399: "two hybrid fragment pooling approach",
    400: "affinity technology",
    401: "biochemical",
    402: "chromatin immunoprecipitation assay",
    403: "colocalization",
    404: "comigration in non denaturing gel electrophoresis",
    405: "competition binding",
    406: "deacetylase assay",
    407: "direct interaction",
    408: "disulfide bond",
    409: "dna footprinting",
    410: "electron tomography",
    411: "enzyme linked immunosorbent assay",
    412: "electrophoretic mobility supershift assay",
    413: "electrophoretic mobility shift assay",
    414: "enzymatic reaction",
    415: "enzymatic study",
    416: "fluorescence microscopy",
    417: "footprinting",
    418: "genetic",
    419: "gtpase assay",
    420: "kinase homogeneous time resolved fluorescence",
    421: "identification by antibody",
    422: "immunostaining",
    423: "in-gel kinase assay",
    424: "protein kinase assay",
    425: "kinase scintillation proximity assay",
    426: "light microscopy",
    427: "Identification by mass spectrometry",
    428: "imaging technique",
    429: "necessary binding region",
    430: "nucleic acid uv cross-linking assay",
    431: "obsolete",
    432: "one hybrid",
    433: "partial identification of protein sequence",
    434: "phosphatase assay",
    435: "protease assay",
    436: "protein footprinting",
    437: "protein three hybrid",
    438: "rna three hybrid",
    439: "random spore analysis",
    440: "saturation binding",
    441: "synthetic genetic analysis",
    442: "sufficient binding region",
    443: "ubiquitin binding",
    444: "database citation",
    445: "literature database",
    446: "pubmed",
    447: "feature database",
    448: "gene ontology",
    449: "interpro",
    450: "cdd",
    451: "pfam",
    452: "pirsf",
    453: "prints",
    454: "prodom",
    455: "prosite",
    456: "scop superfamily",
    457: "smart",
    458: "tigrfams",
    459: "mmdb",
    460: "rcsb pdb",
    461: "interaction database",
    462: "bind",
    463: "biogrid",
    464: "cygd",
    465: "dip",
    466: "ecocyc",
    467: "reactome",
    468: "hprd",
    469: "intact",
    470: "kegg",
    471: "mint",
    472: "pdbe",
    473: "participant database",
    474: "chebi",
    475: "ddbj/embl/genbank",
    476: "ensembl",
    477: "entrez gene/locuslink",
    478: "flybase",
    479: "mgd/mgi",
    480: "omim",
    481: "refseq",
    482: "rfam",
    483: "rgd",
    484: "sgd",
    485: "uniparc",
    486: "uniprot knowledge base",
    487: "wormbase",
    488: "psi-mi",
    489: "source database",
    490: "experiment condition",
    491: "in silico",
    492: "in vitro",
    493: "in vivo",
    494: "in situ",
    495: "experimental role",
    496: "bait",
    497: "neutral component",
    498: "prey",
    499: "unspecified role",
    500: "biological role",
    501: "enzyme",
    502: "enzyme target",
    503: "self",
    505: "experimental feature",
    506: "over expressed level",
    507: "tag",
    508: "deacetylase radiometric assay",
    509: "phosphatase homogeneous time resolved fluorescence",
    510: "homogeneous time resolved fluorescence",
    511: "protease homogeneous time resolved fluorescence",
    512: "zymography",
    513: "collagen film assay",
    514: "in gel phosphatase assay",
    515: "methyltransferase assay",
    516: "methyltransferase radiometric assay",
    517: "radiolabel",
    518: "flag tag",
    519: "glutathione s tranferase tag",
    520: "ha tag",
    521: "his tag",
    522: "myc tag",
    523: "t7 tag",
    524: "calmodulin binding peptide plus protein a tag",
    525: "v5 tag",
    526: "n-acetyl-lysine",
    527: "adp ribosylated residue",
    528: "omega-n-(adp-ribosyl)-arginine",
    529: "s-(adp-ribosyl)-cysteine",
    530: "glutamyl-5-poly(adp-ribose)",
    531: "o-(adp-ribosyl)-serine",
    532: "n4-(adp-ribosyl)-asparagine",
    533: "glycosylated residue",
    534: "glycosyl-cysteine",
    535: "glycosyl-serine",
    536: "glycosyl-threonine",
    537: "omega-n-glycosyl-arginine",
    538: "n4-glycosyl-asparagine",
    539: "gpi anchor residue",
    540: "gpi-anchor amidated alanine",
    541: "gpi-anchor amidated asparagine",
    542: "gpi-anchor amidated aspartate",
    543: "gpi-anchor amidated cysteine",
    544: "gpi-anchor amidated glycine",
    545: "gpi-anchor amidated serine",
    546: "gpi-anchor amidated threonine",
    547: "s-prenyl-cysteine",
    548: "methylated-lysine",
    549: "alkylated cysteine",
    550: "gamma-carboxyglutamic acid",
    551: "nitro-tyrosine",
    552: "s-nitrosyl-cysteine",
    553: "o4'-sulfo-tyrosine",
    554: "sumoylated lysine",
    555: "phospho-histidine",
    556: "transglutamination reaction",
    557: "adp ribosylation reaction",
    558: "deglycosylation reaction",
    559: "glycosylation reaction",
    560: "myristoylated residue",
    561: "palmitoylated residue",
    562: "methylated alanine",
    563: "methylated arginine",
    564: "omega-n-methyl-arginine",
    565: "neddylated lysine",
    566: "sumoylation reaction",
    567: "neddylation reaction",
    568: "desumoylation reaction",
    569: "deneddylation reaction",
    570: "protein cleavage",
    571: "mrna cleavage",
    572: "dna cleavage",
    573: "mutation disrupting interaction",
    574: "digital object identifier",
    575: "alliance for cellular signaling",
    576: "structural proximity",
    577: "feature prediction from structure",
    578: "maltose binding protein tag",
    579: "electron donor",
    580: "electron acceptor",
    581: "suppressor gene",
    582: "suppressed gene",
    583: "fluorescence donor",
    584: "fluorescence acceptor",
    585: "intenz",
    586: "inhibitor",
    587: "inhibited",
    588: "three hybrid",
    589: "in vitro translated protein",
    590: "attribute name",
    591: "experiment description",
    592: "ipfam",
    593: "translocation",
    594: "translocation start",
    595: "translocation end",
    596: "experimental form description",
    597: "feature description",
    598: "feature constraint",
    599: "figure legend",
    600: "conditional synthetic lethal nutrition-sensitivity",
    601: "sequence ontology",
    602: "chemical footprinting",
    603: "dimethylsulphate footprinting",
    604: "potassium permanganate footprinting",
    605: "enzymatic footprinting",
    606: "DNase I footprinting",
    607: "small nuclear rna",
    608: "ribosomal rna",
    609: "small nucleolar rna",
    610: "small interfering rna",
    611: "signal recognition particle rna",
    612: "comment",
    613: "function",
    614: "url",
    615: "search-url",
    616: "example",
    617: "disease",
    618: "caution",
    619: "pathway",
    620: "search-url-ascii",
    621: "author-confidence",
    622: "confidence-mapping",
    623: "inhibition",
    624: "stimulant",
    625: "agonist",
    626: "antagonist",
    627: "experiment modification",
    628: "validation regular expression",
    629: "complex-properties",
    630: "3d-structure",
    631: "3d-r-factors",
    632: "3d-resolution",
    633: "data-processing",
    634: "contact-email",
    635: "contact-comment",
    636: "author-list",
    637: "isoform-comment",
    638: "prerequisite-ptm",
    639: "resulting-ptm",
    640: "parameter type",
    641: "ic50",
    642: "ec50",
    643: "ki",
    644: "km",
    645: "kcat",
    646: "Kd",
    647: "parameter unit",
    648: "molar",
    649: "second",
    650: "millimolar",
    651: "micromolar",
    652: "nanomolar",
    653: "picomolar",
    654: "fentomolar",
    655: "lambda repressor two hybrid",
    656: "identified peptide",
    657: "systematic evolution of ligands by exponential enrichment",
    658: "multidimensional protein identification technology",
    659: "experimental feature detection",
    660: "feature prediction",
    661: "experimental participant identification",
    662: "imex-primary",
    663: "confocal microscopy",
    664: "interaction attribute name",
    665: "experiment attibute name",
    666: "participant attribute name",
    667: "controlled vocabulary attribute name",
    668: "feature attribute name",
    669: "organism attribute name",
    670: "imex",
    671: "antibodies",
    672: "library-used",
    673: "complex synonym",
    674: "peptide parent sequence reference",
    675: "international protein index",
    676: "tandem affinity purification",
    677: "tandem tag",
    678: "antibody array",
    679: "poly adenine",
    680: "single stranded deoxyribonucleic acid",
    681: "double stranded deoxyribonucleic acid",
    682: "cofactor",
    683: "sequence database",
    684: "ancillary",
    685: "source reference",
    686: "unspecified method",
    687: "fluorescent protein tag",
    688: "dna binding domain tag",
    689: "activation domain tag",
    690: "gal4 activation domain",
    691: "vp16 activation domain",
    692: "b42 activation domain",
    693: "gal4 dna binding domain",
    694: "lexa dna binding domain",
    695: "sandwich immunoassay",
    696: "polymerase assay",
    697: "dna directed dna polymerase assay",
    698: "dna directed rna polymerase assay",
    699: "rna directed dna polymerase assay",
    700: "rna directed rna polymerase assay",
    701: "dna strand elongation",
    702: "panther",
    703: "gene3d",
    704: "nucleic acid delivery",
    705: "anti tag western blot",
    706: "nucleic acid transformation",
    707: "anti tag immunostaining",
    708: "monoclonal antibody immunostaining",
    709: "polyclonal antibody immunostaining",
    710: "nucleic acid transformation by treatment with divalent cation",
    711: "nucleic acid electroporation",
    712: "nucleic acid microinjection",
    713: "nucleic acid passive uptake",
    714: "nucleic acid transduction",
    715: "nucleic acid conjugation",
    716: "passive uptake",
    717: "nucleic acid transfection with liposome",
    718: "nucleic acid transfection by treatment",
    719: "calcium phosphate nucleic acid transfection",
    720: "nucleic acid delivery by infection",
    721: "protein delivery",
    722: "protein electroporation",
    723: "protein microinjection",
    724: "protein delivery by cationic lipid treatment",
    725: "protein delivery by infection",
    726: "reverse two hybrid",
    727: "lexa b52 complementation",
    728: "gal4 vp16 complementation",
    729: "luminescence based mammalian interactome mapping",
    730: "pubchem",
    731: "3d repertoire",
    732: "red fluorescent protein tag",
    733: "cyan fluorescent protein tag",
    734: "enhanced green fluorescent protein tag",
    735: "transactivating tag",
    736: "protein passive uptake",
    737: "peptide sequence database",
    738: "pride",
    739: "penetrating tag",
    740: "cell penetrating peptide tag",
    741: "peptide atlas",
    742: "gpm",
    787: "genetic experimental form",
    788: "knock out",
    789: "knock down",
    790: "hypermorph",
    791: "hypomorph",
    792: "antimorph",
    793: "amorph",
    794: "synthetic",
    795: "asynthetic",
    796: "suppression",
    797: "epistasis",
    798: "conditional genetic interaction defined by inequality",
    799: "additive genetic interaction defined by inequality",
    800: "single nonmonotonic genetic interaction defined by inequality",
    801: "double nonmonotonic genetic interaction defined by inequality",
    802: "enhancement interaction",
    803: "expression level alteration",
    804: "mutated gene",
    805: "wwpdb",
    806: "pdbj",
    807: "comigration in gel electrophoresis",
    808: "comigration in sds page",
    809: "bimolecular fluorescence complementation",
    810: "substitution analysis",
    811: "insertion analysis",
    812: "calmodulin binding protein tag",
    813: "proximity ligation assay",
    814: "protease accessibility laddering",
    815: "confirmation by molecular weight",
    816: "molecular weight estimation by staining",
    817: "molecular weight estimation by silver staining",
    818: "molecular weight estimation by coomasie staining",
    819: "molecular weight estimation by bromide staining",
    820: "molecular weight estimation by sybr staining",
    821: "molecular weight estimation by autoradiography",
    822: "molecular weight estimation by hoechst staining",
    823: "predetermined feature",
    824: "x-ray powder diffraction",
    825: "x-ray fiber diffraction",
    826: "x ray scattering",
    827: "x-ray tomography",
    828: "polyprotein fragment",
    829: "multiple parent reference",
    830: "tissue list",
    831: "cell ontology",
    832: "half cystine",
    833: "autoradiography",
    834: "ka",
    835: "koff",
    836: "temperature of interaction",
    837: "pH of interaction",
    838: "kelvin",
    839: "per mole per second",
    840: "stimulator",
    841: "phosphotransferase assay",
    842: "phosphate donor",
    843: "phosphate acceptor",
    844: "phosphotransfer reaction",
    845: "spin label",
    846: "r1 spin label",
    847: "dansyl label",
    848: "125i radiolabel",
    849: "ncbi taxonomy",
    850: "encode",
    851: "protein genbank identifier",
    852: "nucleotide genbank identifier",
    853: "dna overhang",
    854: "3 prime overhang",
    855: "5 prime overhang",
    856: "fluorophore",
    857: "fluorescent dye label",
    858: "immunodepleted coimmunoprecipitation",
    859: "intermolecular force",
    860: "genbank indentifier",
    861: "protein a tag",
    862: "zz tag",
    863: "thiol reactive lanthanide label",
    864: "brenda",
    865: "fluorescence acceptor donor pair",
    866: "tag visualisation",
    867: "tag visualisation by fluorescence",
    868: "author identifier",
    869: "originally assigned identifier",
    870: "demethylase assay",
    871: "demethylation reaction",
    872: "atomic force microscopy",
    873: "curation request",
    875: "dataset",
    878: "author submitted",
    879: "nucleoside triphosphatase assay",
    880: "atpase assay",
    881: "nucleoside triphosphatase reaction",
    882: "atpase reaction",
    883: "gtpase reaction",
    884: "vsv tag",
    885: "journal",
    886: "publication year",
    887: "histone acetylase assay",
    888: "small angle neutron scattering",
    889: "acetylase assay",
    890: "qdot",
    891: "neutron fiber diffraction",
    892: "solid phase assay",
    893: "neutron diffraction",
    894: "electron diffraction",
    895: "protein kinase A complementation",
    896: "renilla luciferase protein tag",
    897: "protein modification ontology",
    898: "putative self",
    899: "p3 filamentous phage display",
    900: "p8 filamentous phage display",
    901: "isotope label footprinting",
    902: "rna cleavage",
    903: "mpidb",
    904: "polysaccharide",
    905: "amplified luminescent proximity homogeneous assay",
    906: "au1 tag",
    907: "conformational status",
    908: "denatured",
    909: "native",
    910: "nucleic acid cleavage",
    911: "cross linker",
    912: "spdp cross linker",
    913: "lc-spdp cross linker",
    914: "association",
    915: "physical association",
    916: "lexa vp16 complementation",
    917: "matrixdb",
    918: "donor",
    919: "acceptor",
    920: "ribonuclease assay",
    921: "surface plasmon resonance array",
    922: "imex evidence",
    923: "irefindex",
    924: "camjedb",
    925: "observed-ptm",
    926: "fiash label",
    927: "iaedans label",
    928: "filter trap assay",
    929: "northern blot",
    930: "epistatis",
    931: "genetic interaction defined by inequality",
    932: "noninteractive",
    933: "negative genetic interaction",
    934: "neutral genetic interaction",
    935: "positive genetic interaction",
    936: "emdb",
    937: "glu tag",
    938: "rheology measurement",
    939: "fluorescein label",
    940: "fluorescein-5-maleimide label",
    941: "competitor",
    942: "uniprot taxonomy",
    943: "detection by mass spectrometry",
    944: "mass spectrometry study of hydrogen/deuterium exchange",
    945: "oxidoreductase activity electron transfer reaction",
    946: "miniaturized immunoprecipitation",
    947: "bead aggregation assay",
    948: "kinetic conditions",
    949: "gdp/gtp exchange assay",
    950: "trapping mutant",
    951: "chain parent sequence reference",
    952: "imex secondary",
    953: "polymerization",
    954: "curation quality",
    955: "curation depth",
    956: "curation coverage",
    957: "full coverage",
    958: "partial coverage",
    959: "imex curation",
    960: "mimix curation",
    961: "rapid curation",
    962: "strep ii tag",
    963: "interactome parallel affinity capture",
    964: "infrared spectroscopy",
    965: "2d-infrared spectrometry",
    966: "ultraviolet-visible spectroscopy",
    967: "chembl compound",
    968: "biosensor",
    969: "bio-layer interferometry",
    970: "inchi key",
    971: "phosphopantetheinylation",
    972: "phosphopantetheinylase assay",
    973: "imex source",
    974: "innatedb",
    975: "fc-igg tag",
    976: "total internal reflection fluorescence spectroscopy",
    977: "no-imex-export",
    978: "author-name",
    979: "oxidoreductase assay",
    980: "tag visualisation by enzyme assay",
    981: "tag visualisation by peroxidase activity",
    982: "electrophoretic mobility-based method",
    983: "gemma",
    984: "deamination assay",
    985: "deamination reaction",
    986: "nucleic acid strand elongation reaction",
    987: "rna strand elongation",
    988: "strep tag",
    989: "amidase assay",
    990: "cleavage assay",
    991: "lipoprotein cleavage assay",
    992: "defarnesylase assay",
    993: "degeranylase assay",
    994: "demyristoylase assay",
    995: "depalmitoylase assay",
    996: "deformylase assay",
    997: "ubiquitinase assay",
    998: "deubiquitinase assay",
    999: "formylase assay",
    1000: "hydroxylase assay",
    1001: "lipidase assay",
    1002: "myristoylase assay",
    1003: "geranylgeranylase assay",
    1004: "palmitoylase assay",
    1005: "adp ribosylase assay",
    1006: "deglycosylase assay",
    1007: "glycosylase assay",
    1008: "sumoylase assay",
    1009: "desumoylase assay",
    1010: "neddylase assay",
    1011: "deneddylase assay",
    1012: "sbp",
    1013: "ensemblgenomes",
    1014: "string",
    1015: "dictybase",
    1016: "fluorescence recovery after photobleaching",
    1017: "rna immunoprecipitation",
    1018: "deltamass",
    1019: "protein phosphatase assay",
    1020: "hilyte fluor 488",
    1021: "qx 520",
    1022: "field flow fractionation",
    1023: "luminogreen",
    1024: "scanning electron microscopy",
    1025: "unimod",
    1026: "diphtamidase assay",
    1027: "diphtamidation reaction",
    1028: "modified chromatin immunoprecipitation",
    1029: "proteomics of isolated chromatin segments",
    1030: "excimer fluorescence",
    1031: "protein folding/unfolding",
    1032: "atto 488",
    1033: "atto 550",
    1034: "nuclease assay",
    1035: "deoxyribonuclease assay",
    1036: "nucleotide exchange assay",
    1037: "Split renilla luciferase complementation",
    1038: "silicon nanowire field-effect transistor",
    1039: "c-terminal range",
    1040: "n-terminal range",
    1041: "synonym",
    1042: "pubmed central",
    1043: "flannotator",
    1044: "rice genome annotation project",
    1045: "curation content",
    1046: "interacting molecules",
    1047: "protein-protein",
    1048: "smallmolecule-protein",
    1049: "nucleicacid-protein",
    1050: "interaction representation",
    1051: "evidence",
    1052: "clustered",
    1053: "data source",
    1054: "experimentally-observed",
    1055: "internally-curated",
    1056: "text-mining",
    1057: "predicted",
    1058: "imported",
    1059: "complex expansion",
    1060: "spoke expansion",
    1061: "matrix expansion",
    1062: "bipartite expansion",
    1063: "consensuspathdb",
    1064: "interaction confidence",
    1065: "replication-based confidence",
    1066: "structure-based confidence",
    1067: "function-based confidence",
    1068: "location-based confidence",
    1069: "network-based confidence",
    1070: "standard-based confidence",
    1071: "literature-based confidence",
    1072: "method-based confidence",
    1073: "statistical-based confidence",
    1074: "rgs-his tag",
    1075: "beilstein",
    1076: "einecs",
    1077: "merck index",
    1078: "plantgdb",
    1079: "ratmap",
    1080: "tair",
    1081: "tigr/jcvi",
    1082: "zfin",
    1083: "cog",
    1084: "photon donor",
    1085: "photon acceptor",
    1086: "equilibrium dialysis",
    1087: "monoclonal antibody blockade",
    1088: "phenotype-based detection assay",
    1089: "nuclear translocation assay",
    1090: "bimane label",
    1091: "publication title",
    1092: "atto label",
    1093: "bibliographic attribute name",
    1094: "genome databases",
    1095: "hgnc",
    1096: "protein sequence databases",
    1097: "uniprot",
    1098: "uniprot/swiss-prot",
    1099: "uniprot/trembl",
    1100: "bioactive entity",
    1101: "standard inchi key",
    1102: "mapped-identity",
    1103: "solution state nmr",
    1104: "solid state nmr",
    1105: "biocyc",
    1106: "pathways database",
    1107: "pid",
    1108: "biocarta",
    1109: "gene database",
    1110: "predicted interaction",
    1111: "two hybrid bait or prey pooling approach",
    1112: "two hybrid prey pooling approach",
    1113: "two hybrid bait and prey pooling approach",
    1114: "virhostnet",
    1115: "spike",
    1116: "genemania",
    1117: "topfind",
    1118: "enhanced yellow fluorescent protein tag",
    1119: "nYFP",
    1120: "cYFP",
    1121: "ceYFP",
    1122: "neYFP",
    1123: "bindingdb",
    1124: "pathwaycommons",
    1125: "direct binding region",
    1126: "self interaction",
    1127: "putative self interaction",
    1128: "mutation disrupting interaction strength",
    1129: "mutation disrupting interaction rate",
    1130: "mutation decreasing interaction rate",
    1131: "mutation increasing interaction rate",
    1132: "mutation increasing interaction strength",
    1133: "mutation decreasing interaction strength",
    1134: "mcherry fluorescent protein tag",
    1135: "venus fluorescent protein tag",
    1136: "kusabira-green protein tag",
    1137: "carboxylation assay",
    1138: "decarboxylation assay",
    1139: "carboxylation reaction",
    1140: "decarboxylation reaction",
    1141: "s tag",
    1142: "aminoacylation assay",
    1143: "aminoacylation reaction",
    1144: "protein a tag visualisation",
    1145: "phospholipase assay",
    1146: "phospholipase reaction",
    1147: "ampylation assay",
    1148: "ampylation reaction",
    1149: "cooperative interaction",
    1150: "affected interaction",
    1151: "participant-ref",
    1152: "cooperative effect value",
    1153: "cooperative effect outcome",
    1154: "positive cooperative effect",
    1155: "negative cooperative effect",
    1156: "cooperative mechanism",
    1157: "allostery",
    1158: "pre-assembly",
    1159: "allosteric molecule",
    1160: "allosteric effector",
    1161: "allosteric response",
    1162: "allosteric k-type response",
    1163: "allosteric v-type response",
    1164: "allosteric mechanism",
    1165: "allosteric change in structure",
    1166: "allosteric change in dynamics",
    1167: "allostery type",
    1168: "heterotropic allostery",
    1169: "homotropic allostery",
    1170: "pre-assembly response",
    1171: "composite binding site formation",
    1172: "altered physicochemical compatibility",
    1173: "binding site hiding",
    1174: "configurational pre-organization",
    1175: "allosteric post-translational modification",
    1176: "sequence based prediction of gene regulatory region binding sites",
    1177: "phylogenetic profiling of predicted gene regulatory region binding sites",
    1178: "sequence based prediction of binding of transcription factor to transcribed gene regulatory elements",
    1179: "partial nucleotide sequence identification",
    1180: "partial DNA sequence identification",
    1181: "paired end tags sequence identification",
    1182: "full identification by RNA sequencing",
    1183: "nuclease footprinting",
    1184: "dna adenine methyltransferase identification",
    1185: "tag visualisation by dna adenine methyltransferase",
    1186: "dna methyltransferase tag",
    1187: "damip",
    1188: "tag visualisation by mutated dna adenine methyltransferase",
    1189: "methylation interference assay",
    1190: "hydroxy radical footprinting",
    1191: "ultraviolet (uv) footprinting",
    1192: "antisense oligonucleotides",
    1193: "partial RNA sequence identification",
    1194: "reverse transcription pcr",
    1195: "quantitative pcr",
    1196: "quantitative reverse transcription pcr",
    1197: "radioimmunoassay",
    1198: "immunohistochemistry",
    1199: "anti-tag immunohistochemistry",
    1200: "immunocytochemistry",
    1201: "anti-tag immunocytochemistry",
    1202: "one-strep-tag",
    1203: "split luciferase complementation",
    1204: "split firefly luciferase complementation",
    1205: "luciferase tag",
    1206: "renilla-n",
    1207: "renilla-c",
    1208: "firefly luciferase protein tag",
    1209: "firefly-c",
    1210: "firefly-n",
    1211: "liposome binding assay",
    1212: "checksum",
    1213: "n-venus",
    1214: "c-venus",
    1215: "bifc tag",
    1216: "cGFP",
    1217: "nGFP",
    1218: "chromosome conformation capture assay",
    1219: "enzyme-mediated activation of radical sources",
    1220: "tag visualisation by luciferase assay",
    1221: "author-based confidence",
    1222: "mbinfo",
    1223: "ptm decreasing an interaction",
    1224: "ptm increasing an interaction",
    1225: "ptm disrupting an interaction",
    1226: "ampylation assay",
    1227: "lap tag",
    1228: "pyo tag",
    1229: "uridylation assay",
    1230: "uridylation reaction",
    1231: "aminomethylcoumarin label",
    1232: "aggregation assay",
    1233: "resulting-cleavage",
    1234: "silac",
    1235: "thermal shift binding",
    1236: "proline isomerase assay",
    1237: "proline isomerization  reaction",
    1238: "mass spectrometry studies of subunit exchange",
    1239: "amino-acid variant",
    1240: "disease causing amino-acid variant",
    1241: "variant",
    1242: "fc-igg1",
    1243: "fc-igg2",
    1244: "mkate2",
    1245: "mkate",
    1246: "ion mobility mass spectrometry of complexes",
    1247: "microscale thermophoresis",
    1248: "bodipy label",
    1249: "isomerase assay",
    1250: "isomerase reaction",
    1251: "methylmalonyl-CoA isomerase reaction",
    1252: "methylmalonyl-CoA isomerase asf say",
    1253: "atto 532",
    1254: "atto 647",
    1255: "stilbene label",
    1256: "luminscent dye label",
    1257: "rhodamine label",
    1258: "tetramethyl rhodamine label",
    1259: "acrylodan label",
    1260: "pyrene label",
    1261: "oregon green label",
    1262: "iid",
    1263: "molecular connections",
    1264: "ntnu",
    1265: "ubiquitin reconstruction tag",
    1266: "cub",
    1267: "nub",
    1268: "nubg",
    1269: "duplicated protein",
    1270: "xpress tag",
    1271: "enhancement",
    1272: "positive epistasis",
    1273: "maximal epistasis",
    1274: "minimal epistasis",
    1275: "neutral epistasis",
    1276: "opposing epistasis",
    1277: "qualitative epistasis",
    1278: "mutual enhancement",
    1279: "unilateral enhancement",
    1280: "mutual suppression",
    1281: "mutual suppression (complete)",
    1282: "mutual suppression (partial)",
    1283: "suppression-enhancement",
    1284: "quantitative epistasis",
    1285: "opposing epistasis",
    1286: "over-suppression",
    1287: "mutual over-suppression",
    1288: "over-suppression-enhancement",
    1289: "phenotype bias",
    1290: "suppression (complete)",
    1291: "suppression (partial)",
    1292: "unilateral suppression",
    1293: "unilateral suppression (complete)",
    1294: "unilateral suppression (partial)",
    1295: "unilateral over-suppression",
    1296: "amino acid analysis",
    1297: "phosphoamino acid analysis",
    1298: "complex type",
    1299: "complex composition",
    1300: "obligate complex",
    1301: "non-obligate complex",
    1302: "stable complex",
    1303: "transient complex",
    1304: "molecule set",
    1305: "candidate set",
    1306: "open set",
    1307: "defined set",
    1308: "resulting sequence",
    1309: "de-ADP-ribosylation assay",
    1310: "de-ADP-ribosylation reaction",
    1311: "differential scanning calorimetry",
    1312: "aut-page",
    1313: "proximity labelling technology",
    1314: "proximity-dependent biotin identification",
    1315: "complex recommended name",
    1316: "complex systematic name",
    1317: "eukaryotic linear motif resource",
    1318: "sulfate donor",
    1319: "sulfate acceptor",
    1320: "membrane yeast two hybrid",
    1321: "ire1 reconstruction",
    1322: "atto 465",
    1323: "tag visualisation by alkaline phosphatase activity",
    1324: "conditioned medium",
    1325: "sulfurtransferase assay",
    1326: "CLONE OF phosphotransfer reaction",
    1327: "sulfurtransfer reaction",
    1328: "coumarin label",
    1329: "cpm",
    1330: "dnp",
    1331: "evidence ontology",
    1332: "bhf-ucl",
    1333: "rogid",
    1334: "rigid",
    1335: "hpidb",
    1336: "experiment database",
    1337: "efo",
    1338: "eef tag",
    1339: "supercharged green fluorescent protein",
    1340: "human orfeome collection",
    1341: "set member",
    1342: "qcmd",
    1343: "enzyme regulator",
    1344: "erythrosin iodoacetamide label",
    1345: "rho tag",
    1346: "bmrb",
    1347: "protein ontology",
    1348: "chembl target",
    1349: "chembl",
    1350: "orphanet",
    1351: "inferred-from",
    1352: "uracil interference assay",
    1353: "au5 tag",
    1354: "lipase assay",
    1355: "lipid cleavage",
    1356: "validated two hybrid",
    1357: "RNAcentral",
    2002: "drugbank",
    2003: "commercial name",
    2004: "drug brand name",
    2005: "drug mixture brand name",
    2006: "biotech product preparation",
    2007: "iupac name",
    2008: "chemical formula",
    2009: "chemical structure",
    2010: "standard inchi",
    2011: "cas registry number",
    2012: "kegg compound",
    2013: "pubchem",
    2015: "pharmgkb",
    2016: "bind smid",
    2017: "heterogen",
    2020: "canadian drug identification number",
    2021: "rxlist link",
    2023: "material safety data sheet",
    2024: "patent number",
    2025: "molecular weight",
    2026: "melting point",
    2027: "water solubility",
    2029: "logp",
    2030: "isoelectric point",
    2033: "hydrophobicity",
    2036: "boiling point",
    2039: "smiles string",
    2040: "drug type",
    2041: "drug category",
    2042: "disease indication",
    2043: "pharmacology",
    2044: "mechanism of action",
    2045: "drug absorption",
    2046: "lethal dose 50",
    2047: "percentage of plasma protein binding",
    2048: "drug biotransformation",
    2049: "elimination half life",
    2050: "dosage form",
    2051: "patient information",
    2053: "contraindications",
    2054: "bioactive entity reference",
    2055: "chemical stability",
    2064: "solubility",
    2084: "organisms affected",
    2086: "physicochemical attribute name",
    2089: "bioactive entity attribute name",
    2091: "structure representation attribute name",
    2097: "anti-convulsant",
    2098: "anti-bacterial",
    2099: "fda approved drug",
    2100: "experimental drug",
    2101: "biotech drug",
    2102: "nutraceutical drug",
    2105: "pka",
    2106: "degree of ionisation ph 7.4",
    2107: "logd",
    2108: "solubility ph 7.4",
    2109: "solubility in dmso",
    2111: "diffusion coefficient",
    2112: "chemical stability at pH 2",
    2113: "dissolution profile",
    2115: "pharmacokinetics attribute name",
    2116: "cell permeability",
    2118: "volume of distribution",
    2120: "tissue distribution",
    2121: "transporter binding",
    2122: "clearance",
    2123: "renal clearance",
    2124: "total clearance",
    2125: "maximum absorbable dose",
    2126: "paracellular absorption",
    2127: "tmax/cmax",
    2128: "ABCB1 transporter substrate",
    2129: "bile transporter substrate",
    2130: "cyp-450 inhibition",
    2131: "metabolite identification",
    2132: "gsh adducts",
    2133: "neutralization by glucuronidation or sulfatation",
    2135: "toxicity attribute name",
    2136: "herg binding",
    2137: "genotoxicity",
    2138: "mutagenicity",
    2139: "carcinogenicity",
    2140: "chromosome damage",
    2141: "hepatotoxicity",
    2142: "phospholipidosis",
    2145: "solubility ph 6.5",
    2146: "solubility ph 2.0",
    2147: "chemical stability at pH 7.4",
    2148: "investigational drug",
    2149: "withdrawn drug",
    2150: "illicit drug",
    2151: "other drug interaction",
    2152: "food interaction",
    2153: "pdr health",
    2154: "wikipedia",
    2155: "average molecular weight",
    2156: "monoisotopic molecular weight",
    2157: "experimental water solubility",
    2158: "predicted water solubility",
    2160: "logs",
    2161: "experimental logs",
    2162: "experimental CaCO2 permeability",
    2163: "by homology",
    2164: "mind",
    2165: "bar",
    2166: "ai",
    2167: "kinetic exclusion assay",
    2168: "conditional site labelling",
    2169: "luminiscence technology",
    2170: "bimolecular luminiscence complementation",
    2171: "complemented donor-acceptor resonance energy transfer",
    2172: "aspgd",
    2173: "cgd",
    2174: "ecoliwiki",
    2175: "genedb",
    2176: "gramene",
    2177: "pombase",
    2178: "agi_locuscode",
    2179: "subset",
    2180: "agbase",
    2181: "cacao",
    2182: "dflat",
    2183: "go_central",
    2184: "mtbbase",
    2185: "parkinsonsuk-ucl",
    2186: "alut",
    2187: "ri",
    2188: "par-clip",
    2189: "avexis",
    2190: "long non-coding ribonucleic acid",
    2191: "clip",
    2192: "clip-seq",
    2193: "iclip",
    2194: "crac",
    2195: "clash",
    2196: "quartz crystal microbalance",
    2197: "probe interaction assay",
    2198: "labelling assay",
    2199: "specific site-labelling technology",
    2200: "primesdb",
    2201: "DNA chemical modification",
    2202: "RNA chemical modification",
    2203: "primer extension assay",
    2204: "micro rna",
    2205: "pir",
    2206: "observed nucleic acid chemical modification",
    2207: "resulting nucleic acid chemical modification",
    2208: "prerequisite-nucleic acid chemical modification",
    2209: "nucleic acid chemical modification decreasing an interaction",
    2210: "nucleic acid chemical modification disrupting an interaction",
    2211: "nucleic acid chemical modification increasing an interaction",
    2212: "proteomexchange",
    2213: "super-resolution microscopy",
    2214: "signor",
    2215: "barcode fusion genetics two hybrid",
    2216: "deampylation assay",
    2217: "luciferase-c",
    2218: "luciferase-n",
    2219: "gaussia luciferase protein tag",
    2220: "gaussia-c",
    2221: "gaussia-n",
    2222: "inference by socio-affinity scoring",
    2223: "inference by quantitative co-purification",
    2224: "chemical rna modification plus base pairing prediction",
    2225: "zinc",
    2226: "mutation with no effect",
    2227: "mutation causing an interaction",
    2228: "ceitec",
    2229: "nucleicacid-gene",
    2230: "nucleicacid-nucleicacid",
    2231: "coexpression",
    2232: "molecular association",
    2233: "causal interaction",
    2234: "causal statement",
    2235: "up-regulates",
    2236: "up-regulates activity",
    2237: "up-regulates quantity",
    2238: "up-regulates quantity by expression",
    2239: "up-regulates quantity by stabilization",
    2240: "down-regulates",
    2241: "down-regulates activity",
    2242: "down-regulates quantity",
    2243: "down-regulates quantity by repression",
    2244: "down-regulates quantity by destabilization",
    2245: "causal regulatory mechanism",
    2246: "indirect causal regulation",
    2247: "transcriptional regulation",
    2248: "translation regulation",
    2249: "post transcriptional regulation",
    2250: "direct causal regulation",
    2251: "transcriptional regulation by direct binding of dbTF to DNA regulatory element",
    2252: "guanine nucleotide exchange factor reaction",
    2253: "gtpase-activating protein reaction",
    2254: "chemical activation reaction",
    2255: "chemical inhibition reaction",
    2256: "relocalization",
    2257: "small molecule catalysis reaction",
    2258: "xenobiotic",
    2259: "causal interactor type",
    2260: "stimulus",
    2261: "phenotype",
    2262: "causal regulatory modification",
    2263: "s-nitrosylation",
    2264: "tyrosinated residue",
    2265: "de-acetylated residue",
    2266: "de-phosphorylated residue",
    2267: "de-sumoylated residue",
    2268: "de-methylated residue",
    2269: "de-ubiquitinylated residue",
    2270: "signalink",
    2271: "edam",
    2272: "tyrosinylation",
    2273: "tyrosination",
    2274: "regulator",
    2275: "regulator target",
    2276: "carbohydrate chemical modification",
    2277: "Cr-two hybrid",
    2278: "polymer chain length",
    2279: "complex portal",
    2280: "deamidation reaction",
    2281: "deamidation assay",
    2282: "complex-primary",
    2283: "southwestern blotting",
    2284: "complex component",
    2285: "miRNA interference luciferase reporter assay",
    2286: "functional association",
    2287: "identification by structure determination",
    2288: "DAP-seq",
}



"""ClinVAR API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestClinvar:
    def test_get_clinvar(self, client):
        response = client.get(
            "api/v1/clinvar?allele_id=705202&type=single%20nucleotide%20variant&name=NM_001772.4%28CD33%29%3Ac.131T%3EC%20%28p.Phe44Ser%29&gene_id=945&gene_symbol=CD33&hgnc_id=HGNC%3A1659&clinical_significance=Benign&last_evaluated=Jul%2016%2C%202018&rs_db_snp=61736469&rcvaccession=RCV000957991&origin=germline&origin_simple=germline&assembly=GRCh37&chromosome_accession=NC_000019.9&chromosome=19&start=51728567&stop=51728567&reference_allele=na&alternate_allele=na&cytogenetic=19q13.41&review_status=criteria%20provided%2C%20single%20submitter&number_submitters=1&tested_in_gtr=N&submitter_categories=2&variation_id=777507&position_vcf=51728567&reference_allele_vcf=T&alternate_allele_vcf=C&page_size=10&page=1",
            content_type="application/json",
        )

        output = format_response_data(response)

        expected_results = {
            "allele_id": 705202,
            "alternate_allele": "na",
            "alternate_allele_vcf": "C",
            "assembly": "GRCh37",
            "chromosome": "19",
            "chromosome_accession": "NC_000019.9",
            "clin_sig_simple": 0,
            "clinical_significance": "Benign",
            "cytogenetic": "19q13.41",
            "gene_id": 945,
            "gene_symbol": "CD33",
            "guidelines": None,
            "hgnc_id": "HGNC:1659",
            "id": 1132137,
            "last_evaluated": "Jul 16, 2018",
            "name": "NM_001772.4(CD33):c.131T>C (p.Phe44Ser)",
            "nsv_esv_db_var": None,
            "number_submitters": 1,
            "origin": "germline",
            "origin_simple": "germline",
            "otherIdentifiers": [],
            "phenotypeMedgens": ["CN517202"],
            "position_vcf": 51728567,
            "rcvaccession": "RCV000957991",
            "reference_allele": "na",
            "reference_allele_vcf": "T",
            "review_status": "criteria provided, single submitter",
            "rs_db_snp": 61736469,
            "start": 51728567,
            "stop": 51728567,
            "submitter_categories": 2,
            "tested_in_gtr": "N",
            "type": "single nucleotide variant",
            "variation_id": 777507,
        }

        assert output[NUM_RESULTS] == 1
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_clinvar_simple(self, client):
        response = client.get(
            "api/v1/clinvar/simple?phenotype=Alzheimer%20disease%202&hgnc_id=HGNC%3A4886&assembly=GRCh38&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "allele_id": 15048,
            "assembly": "GRCh38",
            "gene_symbol": "HFE",
            "hgnc_id": "HGNC:4886",
            "id": 14,
            "rs_db_snp": 1800562,
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_phenotype_starts_with(self, client):
        response = client.get(
            "api/v1/clinvar/phenotype/starts_with?phenotype=MAN1B1-Related%20Disor&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert "MAN1B1-Related Disorders" in results
        assert results["MAN1B1-Related Disorders"] >= 6899  # Value changes depending on when data was downloaded

    def test_get_by_other_identifier(self, client):
        response = client.get(
            "api/v1/clinvar/phenotype/by_other_identifier?db=OMIM&identifier=613653.%25&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "allele_id": 15041,
            "assembly": "GRCh37",
            "db": "OMIM",
            "gene_symbol": "AP5Z1",
            "hgnc_id": "HGNC:22197",
            "id": 1,
            "identifier": "613653.0001",
            "rs_db_snp": 397704705,
        }
        assert output[NUM_RESULTS] > 10
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_by_medgen(self, client):
        response = client.get(
            "api/v1/clinvar/phenotype/by_medgen?identifier=C0002395&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "allele_id": 15048,
            "assembly": "GRCh37",
            "gene_symbol": "HFE",
            "hgnc_id": "HGNC:4886",
            "id": 13,
            "identifier": "C0002395",
            "rs_db_snp": 1800562,
        }
        assert output[NUM_RESULTS] > 300
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""Ensembl."""
import gzip
import re
from typing import Dict

import pandas as pd
from pyorientdb import OrientDB

from ebel.manager.orientdb import odb_meta, urls
from ebel.manager.orientdb.constants import ENSEMBL
from ebel.manager.rdbms.models import ensembl
from ebel.tools import get_file_path

# TODO: use ftp://ftp.ensembl.org/pub/release-102/mysql/homo_sapiens_core_102_38/gene.txt.gz instead CDs


class Ensembl(odb_meta.Graph):
    """Ensembl (https://www.ensembl.org/index.html)."""

    def __init__(self, client: OrientDB = None):
        """Init Ensembl."""
        self.client = client
        self.biodb_name = ENSEMBL
        self.urls = {self.biodb_name: urls.ENSEMBL_CDS}
        super().__init__(urls=self.urls, tables_base=ensembl.Base, biodb_name=self.biodb_name)

    def __len__(self):
        return self.session.query(ensembl.Ensembl).count()

    def __contains__(self, ensembl_id) -> bool:
        count = self.session.query(ensembl.Ensembl).filter(ensembl.Ensembl.gene_id_short == ensembl_id).count()
        return bool(count)

    def insert_data(self) -> Dict[str, int]:
        """Insert Ensembl data."""
        data = []
        file_path = get_file_path(self.urls[self.biodb_name], self.biodb_name)
        with gzip.open(file_path, "r") as f:
            lines = [x.decode("utf-8").strip() for x in f.readlines() if x.startswith(b">")]

        regex = (
            r"^>(?P<enst>ENST\d+)\.\d+ cds chromosome:GRCh"
            r"(?P<version>\d+):"
            r"(?P<chromosome>((1|2)?\d|X|Y|MT)):"
            r"(?P<start>\d+):"
            r"(?P<stop>\d+):"
            r"(?P<orientation>-?1) gene:"
            r"(?P<gene_id>(?P<gene_id_short>ENSG\d+)\.\d+) "
            r"gene_biotype:protein_coding transcript_biotype:protein_coding gene_symbol:"
            r"(?P<symbol>\S+) .*? \[Source:HGNC Symbol;Acc:"
            r"(?P<hgnc_id>HGNC:\d+)\]"
        )
        pattern = re.compile(regex)
        for line in lines:
            found = pattern.search(line)
            if found:
                data += [found.groupdict()]

        df = pd.DataFrame(data)
        df.index += 1
        df.index.rename("id", inplace=True)
        df.to_sql(ensembl.Ensembl.__tablename__, self.engine, if_exists="append")

        self.session.commit()

        return {self.biodb_name: df.shape[0]}

    def update_interactions(self) -> int:
        """Update edges with EnsEMBL metadata."""
        pass



"""Generic BEL API unit tests."""

from .conftest import format_response_data
from .constants import RESULTS


class TestBel:
    """Unit tests for the BEL API module."""

    def test_get_edge(self, client):
        response = client.get(
            "api/v1/bel/edges?subject_node_class=protein&subject_namespace=HGNC&subject_name=IL4&relation=increases&citation_full_journal_name=Molecular%20neurodegeneration&citation_pub_date=2018-06-12&citation_pub_year=2018&citation_last_author=Landreth%20GE&citation_type=PubMed&title=TREM2%20in%20Neurodegenerative%20Diseases.&doi=10.1186%2Fs13024-017-0197-5&object_node_class=protein&object_namespace=HGNC&object_name=TREM2&page_size=10&page=1",
            content_type="application/json",
        )
        cols = [
            "subject_rid",
            "subject_node_class",
            "subject_namespace",
            "subject_name",
            "subject_bel",
            "subject_gene_symbol_involved_in",
            "subject_other_involved_in",
            "edge_rid",
            "relation",
            "evidence",
            "citation_full_journal_name",
            "citation_pub_date",
            "citation_pub_year",
            "citation_last_author",
            "citation_type",
            "author_in_author_list",
            "title",
            "doi",
            "object_rid",
            "object_node_class",
            "object_namespace",
            "object_name",
            "object_bel",
            "object_gene_symbol_involved_in",
            "object_other_involved_in",
        ]

        output = format_response_data(response)
        assert isinstance(output, dict)
        results = output[RESULTS]
        assert isinstance(results, list)
        assert len(results) == 1
        hit = results[0]
        for col in cols:
            assert col in hit
        assert hit["object_name"] == "IL2"
        assert hit["object_namespace"] == "HGNC"
        assert hit["object_node_class"] == "protein"
        assert hit["citation_last_author"] == "Schellenberg GD"
        assert hit["relation"] == "increases"



"""IUPHAR RDBMS model definition."""
from sqlalchemy import (BigInteger, Boolean, Column, ForeignKey, Integer,
                        Numeric, String, Text)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class IupharLigand(Base):
    """Class definition for the iuphar_ligand table."""

    __tablename__ = "iuphar_ligand"
    id = Column(Integer, primary_key=True)

    name = Column(Text)
    species = Column(Text)
    type = Column(Text)
    approved = Column(Boolean)
    withdrawn = Column(Boolean)
    labelled = Column(Boolean)
    radioactive = Column(Boolean)
    pubchem_sid = Column(BigInteger)
    pubchem_cid = Column(Text)  # TODO: This is a integer, but for import reasons this changed to text
    uniprot_id = Column(Text)
    ensembl_id = Column(Text)
    ligand_subunit_ids = Column(Text)
    ligand_subunit_name = Column(Text)
    ligand_subunit_uni_prot_ids = Column(Text)
    ligand_subunit_ensembl_ids = Column(Text)
    iupac_name = Column(Text)
    inn = Column(Text)
    synonyms = Column(Text)
    smiles = Column(Text)
    inchi_key = Column(Text)
    inchi = Column(Text)
    gto_immu_pdb = Column(Boolean)
    gto_mpdb = Column(Boolean)
    antibacterial = Column(Boolean)

    interactions = relationship("IupharInteraction")

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)


class IupharInteraction(Base):
    """Class definition for the iuphar_interaction table."""

    __tablename__ = "iuphar_interaction"
    id = Column(Integer, primary_key=True)

    target = Column(String(255))
    target_id = Column(Integer)
    target_subunit_ids = Column(Text)
    target_gene_symbol = Column(String(100))
    target_uniprot = Column(String(100))
    target_ensembl_gene_id = Column(String(200))
    target_ligand = Column(String(100))
    target_ligand_id = Column(Integer)
    target_ligand_subunit_ids = Column(Text)
    target_ligand_gene_symbol = Column(String(50))
    target_ligand_uniprot_id = Column(String(200))
    target_ligand_ensembl_gene_id = Column(String(50))
    target_ligand_pubchem_sid = Column(Integer)
    target_species = Column(String(100))
    ligand = Column(String(255))
    ligand_id = Column(Integer, ForeignKey("iuphar_ligand.id"), index=True)
    ligand_subunit_ids = Column(Text)
    ligand_gene_symbol = Column(String(50))
    ligand_species = Column(String(50))
    ligand_pubchem_sid = Column(Integer)
    ligand_type = Column(Text)
    approved = Column(Boolean)
    type = Column(String(100))
    action = Column(String(100))
    action_comment = Column(String(255))
    selectivity = Column(String(50))
    endogenous = Column(Boolean)
    primary_target = Column(Boolean)
    concentration_range = Column(String(50))
    affinity_units = Column(String(10))
    affinity_high = Column(Numeric(6, 2))
    affinity_median = Column(Numeric(6, 2))
    affinity_low = Column(Numeric(6, 2))
    original_affinity_units = Column(String(10))
    original_affinity_low_nm = Column(Numeric(12, 3))
    original_affinity_median_nm = Column(Numeric(12, 3))
    original_affinity_high_nm = Column(Numeric(12, 3))
    original_affinity_relation = Column(String(1))
    assay_description = Column(Text)
    receptor_site = Column(String(100))
    ligand_context = Column(String(50))
    pubmed_id = Column(Text)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)



"""Reactome."""
import json
from typing import Dict

import pandas as pd
from pyorientdb import OrientDB
from sqlalchemy import distinct
from tqdm import tqdm

from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.constants import REACTOME
from ebel.manager.rdbms.models import reactome
from ebel.tools import get_file_path


class Reactome(odb_meta.Graph):
    """Reactome."""

    def __init__(self, client: OrientDB = None):
        """Reactome database."""
        self.client = client
        self.biodb_name = REACTOME
        self.urls = {self.biodb_name: urls.REACTOME}
        super().__init__(
            nodes=odb_structure.reactome_nodes,
            tables_base=reactome.Base,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )

    def __repr__(self):
        """Represent the class."""
        template = "{{BioDatabase:{class_name}}}[url:{url}, nodes_with_reactome:{num_reactome}, generics:{generics}]"
        num = self.query_get_dict("Select count(*) from V where reactome_pathways is not null")[0]["count"]
        representation = template.format(
            class_name=self.__class__.__name__,
            url=self.urls,
            num_reactome=num,
            generics={k: v for k, v in self.number_of_generics.items() if v},
        )
        return representation

    def __len__(self) -> int:
        return self.number_of_nodes

    def __contains__(self, reactome_id: str) -> bool:
        return bool(len(self.execute("Select 1 from reactome where id = '{}' limit 1".format(reactome_id))))

    def insert_data(self) -> Dict[str, int]:
        """Insert data into OrientDB database."""
        columns = [
            "uniprot_accession",
            "identifier",
            "url",
            "name",
            "evidence_type",
            "organism",
        ]
        # evidence_type are:
        # 1. IEA:= Inferred from Electronic Annotation
        # 2. TAS:= Traceable Author Statement
        df = pd.read_csv(
            get_file_path(self.urls[self.biodb_name], self.biodb_name),
            sep="\t",
            names=columns,
            usecols=[0, 1, 3, 4, 5],
        )
        df.index += 1
        df.index.rename("id", inplace=True)
        df.name = df.name.str.strip()
        df.to_sql(reactome.Reactome.__tablename__, self.engine, if_exists="append")

        self.session.commit()

        return {self.biodb_name: df.shape[0]}

    def update_interactions(self) -> int:
        """Update all protein.reactome where protein is pure."""
        self.execute("Update protein set reactome_pathways = null")
        sql_update = "Update protein set reactome_pathways = {} where uniprot = '{}' and pure=true"

        sql_accessions = (
            "Select distinct(uniprot) as accession_id from protein " "where uniprot IS NOT NULL and pure=true"
        )

        proteins = self.execute(sql_accessions)

        updated = 0
        if proteins:
            for unipid_acc in tqdm(proteins, desc="Update Reactome info in pure proteins."):
                accession_id = unipid_acc.oRecordData["accession_id"]
                results = (
                    self.session.query(distinct(reactome.Reactome.name))
                    .filter(reactome.Reactome.uniprot_accession == accession_id)
                    .all()
                )
                if results:
                    reactome_pathways = json.dumps([x[0] for x in results])
                    updated += self.execute(sql_update.format(reactome_pathways, accession_id))[0]

        return updated



"""DisGeNet API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, RESULTS


class TestDisgenet:
    def test_get_sources(self, client):
        response = client.get("api/v1/disgenet/sources", content_type="application/json")
        output = format_response_data(response)
        expected_sources = (
            "BEFREE",
            "CGI",
            "CLINGEN",
            "CLINVAR",
            "CTD_human",
            "CTD_mouse",
            "CTD_rat",
            "GENOMICS_ENGLAND",
            "GWASCAT",
            "GWASDB",
            "HPO",
            "LHGDN",
            "MGD",
            "ORPHANET",
            "PSYGENET",
            "RGD",
            "UNIPROT",
        )
        assert isinstance(output, dict)
        assert all([col in output for col in expected_sources])
        assert all([isinstance(val, int) for val in output.values()])

    def test_get_disease_name_starts_with(self, client):
        response = client.get(
            "api/v1/disgenet/disease_name/starts_with?disease_name=alz&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "ALZHEIMER DISEASE 10": "C1864828",
            "ALZHEIMER DISEASE 18": "C3810041",
            "ALZHEIMER DISEASE 19": "C3810349",
            "ALZHEIMER DISEASE 2": "C1863051",
            "ALZHEIMER DISEASE 4": "C1847200",
            "ALZHEIMER DISEASE 5": "C1865868",
            "ALZHEIMER DISEASE 6, LATE-ONSET": "C1854187",
            "Alzheimer Disease 12": "C1970209",
            "Alzheimer Disease 14": "C1970144",
            "Alzheimer Disease 7": "C1853555",
        }
        assert output[NUM_RESULTS] >= 26
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert results == expected_results

    def test_get_gene_symbol_starts_with(self, client):
        response = client.get(
            "api/v1/disgenet/gene_symbol/starts_with?gene_symbol=CD18&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {"CD180": 4064}
        assert output[NUM_RESULTS] == 1
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert results == expected_results

    def test_get_gene_disease_pmid_associations(self, client):
        response = client.get(
            "api/v1/disgenet/gene_disease_pmid_associations?gene_id=945&gene_symbol=CD33&disease_id=C0023418&disease_name=leukemia&pmid=15388576&source=LHGDN&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "disease_id": "C0023418",
            "disease_name": "leukemia",
            "gene_id": 945,
            "gene_symbol": "CD33",
            "pmid": 15388576,
            "score": 0.1,
            "source": "LHGDN",
        }
        assert output[NUM_RESULTS] == 1
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_variant_disease_pmid_associations(self, client):
        response = client.get(
            "api/v1/disgenet/variant_disease_pmid_associations?snp_id=rs1001179&chromosome=11&position=34438684&disease_id=C0002395&disease_name=Alzheimer%27s%20Disease&pmid=18248894&source=BEFREE&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "chromosome": "11",
            "disease_id": "C0002395",
            "disease_name": "Alzheimer's Disease",
            "pmid": 18248894,
            "position": 34438684,
            "score": 0.01,
            "snp_id": "rs1001179",
            "source": "BEFREE",
        }
        assert output[NUM_RESULTS] == 1
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_ebel_has_snp_disgenet(self, client):
        response = client.get(
            "api/v1/disgenet/ebel?disease_name=Alzheimer%27s%20Disease&rs_number=rs10408847&pmid=29777097&source=GWASCAT&gene_symbol=MARK4&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expect_cols = ("disease_name", "gene_symbol", "pmids", "relation", "rs_number", "score", "source")
        assert output[NUM_RESULTS] > 0
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expect_cols])



"""CHEBI RDBMS model definition."""

from sqlalchemy import (Column, DateTime, ForeignKey, Index, Integer, String,
                        Text)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()


class ChemicalData(Base):
    """Class definition for the chebi_chemical_data table."""

    __tablename__ = "chebi_chemical_data"
    id = Column(Integer, primary_key=True)

    chemical_data = Column(Text, nullable=True)
    source = Column(Text, nullable=False)
    type = Column(Text, nullable=False)

    compound_id = Column(Integer, ForeignKey("chebi_compound.id"))
    compounds = relationship("Compound", back_populates="chemicalData")

    def __str__(self):
        """Class string definition."""
        return self.chemical_data

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "chemical_data": self.chemical_data,
            "source": self.source,
            "type": self.type,
        }


class Comment(Base):
    """Class definition for the chebi_comment table."""

    __tablename__ = "chebi_comment"
    id = Column(Integer, primary_key=True)

    text = Column(Text, nullable=False)
    created_on = Column(DateTime, nullable=False)
    datatype = Column(String(80))
    datatype_id = Column(Integer, nullable=False)

    compound_id = Column(Integer, ForeignKey("chebi_compound.id"))
    compounds = relationship("Compound", back_populates="comments")

    def __str__(self):
        """Class string definition."""
        return self.text

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "text": self.text,
            "datatype": self.datatype,
        }


class Compound(Base):
    """Class definition for the chebi_compound table."""

    __tablename__ = "chebi_compound"
    id = Column(Integer, primary_key=True)

    name = Column(String(2000))
    source = Column(String(32), nullable=False)
    parent_id = Column(Integer)
    chebi_accession = Column(String(30), nullable=False)
    status = Column(String(1), nullable=False)
    definition = Column(Text)
    star = Column(Integer, nullable=False)
    modified_on = Column(Text)
    created_by = Column(Text)

    chemicalData = relationship("ChemicalData", back_populates="compounds")
    comments = relationship("Comment", back_populates="compounds")
    database_accessions = relationship("DatabaseAccession", back_populates="compounds")
    names = relationship("Name", back_populates="compounds")
    references = relationship("Reference", back_populates="compounds")
    # final_id_relations = relationship("Relation", back_populates="final_id_compounds")
    # init_id_relations = relationship("Relation", back_populates="init_id_compounds")
    structures = relationship("Structure", back_populates="compounds")
    inchis = relationship("Inchi", back_populates="compounds")

    def __str__(self):
        return self.name

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "name": self.name,
            "source": self.source,
            "parent_id": self.parent_id,
            "chebi_accession": self.chebi_accession,
            "status": self.status,
            "definition": self.definition,
            "chemicalData": [x.as_dict() for x in self.chemicalData],
            "comments": [x.as_dict() for x in self.comments],
            "database_accessions": [x.as_dict() for x in self.database_accessions],
            "names": [x.as_dict() for x in self.names],
            "references": [x.as_dict() for x in self.references],
            "inchis": [x.as_dict() for x in self.inchis],
        }


class Inchi(Base):
    """Class definition for the chebi_inchi table."""

    __tablename__ = "chebi_inchi"
    id = Column(Integer, primary_key=True)

    inchi = Column(Text)

    compound_id = Column(Integer, ForeignKey("chebi_compound.id"))
    compounds = relationship("Compound", back_populates="inchis")

    def __str__(self):
        return self.inchi

    def as_dict(self):
        """Convert object values to dictionary."""
        return {"inchi": self.inchi}


class DatabaseAccession(Base):
    """Class definition for the chebi_database_accession table."""

    __tablename__ = "chebi_database_accession"
    id = Column(Integer, primary_key=True)

    accession_number = Column(String(255), nullable=True)
    type = Column(Text, nullable=False)
    source = Column(Text, nullable=False)

    compound_id = Column(Integer, ForeignKey("chebi_compound.id"))
    compounds = relationship("Compound", back_populates="database_accessions")

    def __str__(self):
        return self.accession_number

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "accession_number": self.accession_number,
            "type": self.type,
            "source": self.source,
        }


class Name(Base):
    """Class definition for the chebi_name table."""

    __tablename__ = "chebi_name"
    id = Column(Integer, primary_key=True)

    name = Column(Text, nullable=True)
    type = Column(Text, nullable=False)
    source = Column(Text, nullable=False)
    adapted = Column(Text, nullable=False)
    language = Column(Text, nullable=False)

    compound_id = Column(Integer, ForeignKey("chebi_compound.id"))
    compounds = relationship("Compound", back_populates="names")

    def __str__(self):
        return self.name

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "name": self.name,
            "type": self.type,
            "source": self.source,
            "adapted": self.adapted,
            "language": self.language,
        }


class Reference(Base):
    """Class definition for the chebi_reference table."""

    __tablename__ = "chebi_reference"

    id = Column(Integer, primary_key=True)

    reference_id = Column(String(60), nullable=False, index=True)
    reference_db_name = Column(String(60), nullable=False, index=True)
    location_in_ref = Column(String(90), index=True)
    reference_name = Column(String(1024))

    compound_id = Column(Integer, ForeignKey("chebi_compound.id"))
    compounds = relationship("Compound", back_populates="references")

    __table_args__ = (Index("ix_chebi_reference__reference_name", reference_name, mysql_length=500),)

    def __str__(self):
        return f"{self.reference_db_name}:{self.reference_id}"

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "reference_id": self.reference_id,
            "reference_db_name": self.reference_db_name,
            "location_in_ref": self.location_in_ref,
            "reference_name": self.reference_name,
        }

    def as_dict_with_compound_id(self):
        """Convert object values to dictionary with compound ID."""
        return {
            "reference_id": self.reference_id,
            "reference_db_name": self.reference_db_name,
            "location_in_ref": self.location_in_ref,
            "reference_name": self.reference_name,
            "compound_id": self.compound_id,
        }


class Relation(Base):
    """Class definition for the chebi_relation table."""

    __tablename__ = "chebi_relation"
    id = Column(Integer, primary_key=True)

    type = Column(Text, nullable=False)
    status = Column(String(1), nullable=False)

    final_id = Column(Integer, ForeignKey("chebi_compound.id"))
    init_id = Column(Integer, ForeignKey("chebi_compound.id"))

    final_id_compounds = relationship("Compound", foreign_keys=[final_id])
    init_id_compounds = relationship("Compound", foreign_keys=[init_id])

    def __str__(self):
        return f"{self.type} - {self.status}"

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "type": self.type,
            "status": self.status,
            "final_id": self.final_id,
            "init_id": self.init_id,
        }


class Structure(Base):
    """Class definition for the chebi_structure table."""

    __tablename__ = "chebi_structure"
    id = Column(Integer, primary_key=True)

    structure = Column(Text, nullable=False)
    type = Column(Text, nullable=False)
    dimension = Column(Text, nullable=False)
    default_structure = Column(String(1), nullable=False)
    autogen_structure = Column(String(1), nullable=False)

    compound_id = Column(Integer, ForeignKey("chebi_compound.id"))
    compounds = relationship("Compound", back_populates="structures")

    def __str__(self):
        return self.structure



"""DrugBank API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestDrugbank:
    def test_get_by_id(self, client):
        response = client.get("api/v1/drugbank/by_id?drugbank_id=DB00843", content_type="application/json")
        output = format_response_data(response)
        expected_results = {  # Only check a subset
            "cas_number": "120014-06-4",
            "drugbank_id": "DB00843",
            "name": "Donepezil",
        }

        assert isinstance(output, dict)
        assert all([output[key] == value for key, value in expected_results.items()])

    def test_get_drugbank(self, client):
        response = client.get(
            "api/v1/drugbank?drugbank_id=DB00843&name=Donepezil&description=%25Aricept%25&cas_number=120014-06-4&unii=8SSC91326P&state=solid&indication=%25mild%20to%20moderate%20Alzheimer%E2%80%99s%20Disease%25&pharmacodynamics=%25inhibiting%25&toxicity=%25rat%20oral%20LD50%25&metabolism=%25CYP3A4%25&absorption=%25gastrointestinal%20tract%25&half_life=%25hours%25&route_of_elimination=%25urine%25&volume_of_distribution=%25mg%20dose%25&clearance=%25plasma%25&mechanism_of_action=%25cognitive%20and%20behavioral%20decline%25&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {  # Only check a subset
            "cas_number": "120014-06-4",
            "drugbank_id": "DB00843",
            "name": "Donepezil",
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([hit[key] == value for key, value in expected_results.items()])

    def test_get_interaction(self, client):
        response = client.get(
            "api/v1/drugbank/interaction?drugbank_id=DB06605&name=Apixaban&description=%25hemorrhage%25&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_cols = ("description", "drugbank_id", "interactor_drugbank_id", "name")
        assert output[NUM_RESULTS] >= 150
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])
        assert hit["name"] == "Apixaban"

    def test_get_pathway(self, client):
        response = client.get(
            "api/v1/drugbank/pathway?drugbank_id=DB00114&page_size=10&page=1", content_type="application/json"
        )
        output = format_response_data(response)
        expected_cols = ("drugbank_id", "smpdb_id")
        assert output[NUM_RESULTS] == 30
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])
        assert hit["drugbank_id"] == "DB00114"

    def test_get_status(self, client):
        response = client.get(
            "api/v1/drugbank/status?drugbank_id=DB06605&page_size=10&page=1", content_type="application/json"
        )
        output = format_response_data(response)
        expected_results = {"drugbank_id": "DB06605", "smpdb_id": "approved"}
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_patent(self, client):
        response = client.get(
            "api/v1/drugbank/patent?drugbank_id=DB00843&page_size=10&page=1", content_type="application/json"
        )
        output = format_response_data(response)
        expected_cols = ("approved", "country", "drugbank_id", "expires", "number", "pediatric_extension")
        assert output[NUM_RESULTS] == 19
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])
        assert hit["drugbank_id"] == "DB00843"

    def test_get_external_identifier(self, client):
        response = client.get(
            "api/v1/drugbank/external_identifier?drugbank_id=DB00114&resource=BindingDB&identifier=50118216&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {"drugbank_id": "DB00114", "identifier": "50118216", "resource": "BindingDB"}
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_reference(self, client):
        response = client.get(
            "api/v1/drugbank/reference?drugbank_id=DB06605&page_size=10&page=1", content_type="application/json"
        )
        output = format_response_data(response)
        expected_cols = ("drugbank_id", "pmid")
        assert output[NUM_RESULTS] == 2
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])
        assert hit["drugbank_id"] == "DB06605"
        assert hit["pmid"] == 18832478

    def test_get_target(self, client):
        response = client.get(
            "api/v1/drugbank/target?drugbank_id=DB06605&page_size=10&page=1", content_type="application/json"
        )
        output = format_response_data(response)
        expected_results = {
            "action": "inhibitor",
            "drugbank_id": "DB06605",
            "known_action": "yes",
            "uniprot": "P00742",
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_product_name(self, client):
        response = client.get(
            "api/v1/drugbank/product_name?drugbank_id=DB06605&page_size=10&page=1", content_type="application/json"
        )
        output = format_response_data(response)
        expected_cols = ("drugbank_id", "name")
        assert output[NUM_RESULTS] == 3
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])
        assert hit["drugbank_id"] == "DB06605"

    def test_get_synonym(self, client):
        response = client.get(
            "api/v1/drugbank/synonym?drugbank_id=DB06605&page_size=10&page=1", content_type="application/json"
        )
        output = format_response_data(response)
        expected_cols = ("drugbank_id", "synonym")
        assert output[NUM_RESULTS] == 3
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])
        assert hit["drugbank_id"] == "DB06605"
        synonyms = set([x["synonym"] for x in results])
        assert all([syn in synonyms for syn in ("Apixaban", "apixabán", "apixabanum")])



"""NCBI module."""
from typing import Dict, Optional

import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.manager.orientdb import odb_meta, urls
from ebel.manager.orientdb.constants import NCBI
from ebel.manager.rdbms.models import ncbi
from ebel.tools import get_file_path, get_standard_name


class Ncbi(odb_meta.Graph):
    """NCBI Gene (https://ftp.ncbi.nih.gov/gene/DATA/)."""

    def __init__(self, client: OrientDB = None):
        """Init NcbiGene."""
        self.client = client
        self.biodb_name = NCBI
        self.urls = {
            ncbi.NcbiGeneInfo.__tablename__: urls.NCBI_GENE_INFO,
            ncbi.NcbiGeneMim.__tablename__: urls.NCBI_GENE_2MIM,
            ncbi.NcbiGeneEnsembl.__tablename__: urls.NCBI_GENE_2ENSEMBL,
            ncbi.NcbiGeneGo.__tablename__: urls.NCBI_GENE_2GO,
            ncbi.NcbiGenePubmed.__tablename__: urls.NCBI_GENE_2PUBMED,
            ncbi.NcbiGeneOrtholog.__tablename__: urls.NCBI_GENE_ORTHOLOG,
            ncbi.NcbiMedGenName.__tablename__: urls.NCBI_GENE_MEDGEN_NAMES,
            ncbi.NcbiMedGenPmid.__tablename__: urls.NCBI_GENE_MEDGEN_PUBMED,
            "neighbors": urls.NCBI_GENE_NEIGHBORS,
        }
        super().__init__(urls=self.urls, tables_base=ncbi.Base, biodb_name=self.biodb_name)

    def __len__(self):
        return self.session.query(ncbi.NcbiGeneInfo).count()

    def __contains__(self, gene_id) -> bool:
        count = self.session.query(ncbi.NcbiGeneInfo).filter(ncbi.NcbiGeneInfo.gene_id == gene_id).count()
        return bool(count)

    def insert_data(self) -> Dict[str, int]:
        """Insert NCBI gene data.

        :return Dict[str, int]: Tables as keys and inserted rows and values.
        """
        self.recreate_tables()
        inserts = dict()
        inserts[ncbi.NcbiGeneInfo.__tablename__] = self._insert_info()
        inserts[ncbi.NcbiGeneGo.__tablename__] = self._insert_go()
        inserts[ncbi.NcbiGeneEnsembl.__tablename__] = self._insert_ensembl()
        # inserts[ncbi.NcbiGeneMim.__tablename__] = self._insert_mim()
        inserts[ncbi.NcbiGenePubmed.__tablename__] = self._insert_pubmeds()
        inserts[ncbi.NcbiGeneOrtholog.__tablename__] = self._insert_orthologs()
        inserts.update(self._insert_neighbors())
        inserts[ncbi.NcbiMedGenName.__tablename__] = self._insert_medgen()
        return inserts

    def __insert_table(self, model, use_cols=None, sep: str = "\t", chunksize: Optional[int] = None) -> int:
        """Generic method to insert data.

        :param model: NCBI SQLAlchemy model
        :param use_cols: Column names to be used in file
        :param sep: Column separator
        :return:
        """
        table = model.__tablename__
        file_path = get_file_path(self.urls[table], self.biodb_name)
        dfs = pd.read_csv(file_path, sep=sep, usecols=use_cols, chunksize=chunksize)
        if chunksize is None:
            dfs = [dfs]

        number_of_inserts = 0

        for df in dfs:
            self._standardize_dataframe(df)
            df.to_sql(table, self.engine, if_exists="append", index=False)
            number_of_inserts += df.shape[0]

        return number_of_inserts

    def _insert_medgen(self) -> int:
        """Insert MedGen names.

        :return: Number of inserts
        """
        table_name = ncbi.NcbiMedGenName.__tablename__
        file_path_name = get_file_path(self.urls[table_name], self.biodb_name)
        use_cols_name = ["#CUI", "name", "source", "SUPPRESS"]
        df_name = pd.read_csv(file_path_name, usecols=use_cols_name, sep="|").rename_axis("id")
        df_name.index += 1
        self._standardize_dataframe(df_name)
        df_name.to_sql(table_name, self.engine, if_exists="append")

        table_pmid = ncbi.NcbiMedGenPmid.__tablename__
        file_path_pmid = get_file_path(self.urls[table_pmid], self.biodb_name)
        use_cols_pmid = ["CUI", "PMID"]
        dfs_pmid = pd.read_csv(
            file_path_pmid,
            index_col="CUI",
            usecols=use_cols_pmid,
            sep="|",
            chunksize=1000000,
        )

        df_name["ncbi_medgen_name_id"] = df_name.index
        df_name_fk = df_name.set_index("cui")[["ncbi_medgen_name_id"]]

        inserted = 0
        for df in dfs_pmid:
            df.columns = [get_standard_name(x) for x in df.columns]
            df_join = df.join(df_name_fk).reset_index(drop=True)
            inserted += df_join.shape[0]
            df_join.to_sql(table_pmid, self.engine, if_exists="append", index=False)

        return inserted

    def _insert_go(self) -> int:
        """Insert Gene Ontology mapping to Gene IDs.

        :return: Number of inserts
        """
        table = ncbi.NcbiGeneGo.__tablename__
        file_path = get_file_path(self.urls[table], self.biodb_name)
        df = pd.read_csv(file_path, sep="\t")

        table_pmid = ncbi.NcbiGeneGoPmid.__tablename__

        self._standardize_dataframe(df)
        df.index += 1
        df.index.rename("id", inplace=True)
        df.pub_med = df.pub_med.str.split("|")
        rename_columns = {"id": "ncbi_gene_go_id", "pub_med": "pmid"}
        df_pub_med = df[["pub_med"]].explode("pub_med").dropna().reset_index().rename(columns=rename_columns)

        df.drop(columns=["pub_med"], inplace=True)
        df.to_sql(table, self.engine, if_exists="append")
        df_pub_med.to_sql(table_pmid, self.engine, if_exists="append", index=False)

        return df.shape[0]

    def _insert_ensembl(self) -> int:
        """Insert Ensembl mapping to Gene IDs.

        :return: Number of inserts
        """
        return self.__insert_table(ncbi.NcbiGeneEnsembl)

    def _insert_mim(self) -> int:
        """Insert OMIM (Online Mendelian Inheritance in Man) mapping to Gene IDs.

        :return: Number of inserts
        """
        return self.__insert_table(ncbi.NcbiGeneMim)

    def _insert_pubmeds(self) -> int:
        """Insert PubMed references to Gene IDs.

        :return: Number of inserts
        """
        return self.__insert_table(ncbi.NcbiGenePubmed, chunksize=1000000)

    def _insert_orthologs(self) -> int:
        """Insert orthologs to Gene IDs.

        :return: Number of inserts
        """
        use_cols = ["#tax_id", "GeneID", "Other_tax_id", "Other_GeneID"]
        return self.__insert_table(ncbi.NcbiGeneOrtholog, use_cols, chunksize=1000000)

    def _insert_neighbors(self) -> Dict[str, int]:
        """Insert overlapping, right and left neighbors.

        :return int: Number of inserts
        """
        usecols = [
            "GeneID",
            "GeneIDs_on_left",
            "GeneIDs_on_right",
            "overlapping_GeneIDs",
        ]
        file_path = get_file_path(self.urls["neighbors"], self.biodb_name)

        neighbor_types = {
            "overlapping_gene_ids": ncbi.NcbiGeneOverlapping,
            "gene_ids_on_right": ncbi.NcbiGeneOnRight,
            "gene_ids_on_left": ncbi.NcbiGeneOnLeft,
        }

        inserted = {neighbor_type: 0 for neighbor_type in neighbor_types}

        dfs = pd.read_csv(file_path, sep="\t", usecols=usecols, chunksize=1000000)

        for df in dfs:
            df.columns = [get_standard_name(x) for x in df.columns]

            for neighbor_type, model in neighbor_types.items():
                df_type = df[["gene_id", neighbor_type]].set_index("gene_id")
                ntype = neighbor_type.replace("ids", "id")
                df_type.rename(columns={neighbor_type: ntype}, inplace=True)
                df_type[ntype] = df_type[ntype].str.split("|")
                df_type = df_type.explode(ntype)
                df_type = df_type[df_type[ntype] != "-"].reset_index()
                df_type.to_sql(model.__tablename__, self.engine, if_exists="append", index=False)
                inserted[neighbor_type] += df_type.shape[0]
        return inserted

    def _insert_info_description(self) -> pd.DataFrame:
        """Insert gene description.

        :return int: Number of inserts
        """
        table = ncbi.NcbiGeneInfo.__tablename__
        file_path = get_file_path(self.urls[table], self.biodb_name)
        df = (
            pd.read_csv(
                file_path,
                sep="\t",
                usecols=["description"],
                dtype={"description": "string"},
            )
            .dropna()
            .drop_duplicates()
        )
        df.index += 1
        df.index.rename("id", inplace=True)
        df.to_sql(ncbi.NcbiGeneInfoDescription.__tablename__, self.engine, if_exists="append")
        return df.assign(description_id=df.index).set_index(["description"])

    def _insert_info_xref(self, dataframe: pd.DataFrame) -> int:
        """Insert cross references to Gene IDs using Gene Info DataFrame.

        :param dataframe: Gene info pandas DataFrame
        :return int: Number of inserts
        """
        inserted = 0
        df = dataframe[["db_xrefs", "gene_id"]][dataframe.db_xrefs != "-"].dropna()

        if not df.empty:
            df.set_index("gene_id", inplace=True)
            df["db_xrefs"] = df["db_xrefs"].str.split("|")
            df = df.explode("db_xrefs")
            df[["db", "dbid"]] = df["db_xrefs"].str.split(":", n=1, expand=True)
            df.drop(columns=["db_xrefs"], inplace=True)
            df.reset_index().to_sql("ncbi_gene_info_xref", self.engine, if_exists="append", index=False)
            inserted = df.shape[0]

        return inserted

    def _insert_info(self, chunksize: int = 1000000) -> int:
        """Insert Gene Info.

        :param chunksize: size of the row aggregates
        :return int: Number of inserts
        """
        df_info_decr = self._insert_info_description()
        inserted = 0
        table = ncbi.NcbiGeneInfo.__tablename__
        file_path = get_file_path(self.urls[table], self.biodb_name)

        # TODO: To decrease the size of gene_info table, type_of_gene should be implemented as its own table
        # df_type_of_gene = pd.read_csv(file_path, sep="\t",usecols=['type_of_gene']).drop_duplicates()\
        #     .reset_index(drop=True).rename_axis('id')
        # df_type_of_gene.index += 1
        # df_type_of_gene.to_sql()

        use_cols = {
            "#tax_id",
            "GeneID",
            "LocusTag",
            "Symbol",
            "chromosome",
            "description",
            "map_location",
            "type_of_gene",
            "dbXrefs",
        }
        for df in tqdm(
            pd.read_csv(
                file_path,
                sep="\t",
                low_memory=False,
                usecols=use_cols,
                chunksize=chunksize,
            ),
            desc=f"Import {self.biodb_name.upper()}",
        ):
            self._standardize_dataframe(df)

            df.drop(columns=["db_xrefs"]).set_index(["description"]).join(df_info_decr).set_index("gene_id").to_sql(
                table, self.engine, if_exists="append", index=True
            )
            inserted += df.shape[0]

            self._insert_info_xref(df)

        return inserted

    def update_interactions(self) -> int:
        """Abstract method."""
        pass



"""Methods for interfacing with the RDBMS."""
import getpass
import logging
import sys
from typing import Optional, Union

import pymysql
from pyorientdb import OrientDB
from pyorientdb.exceptions import (PyOrientCommandException,
                                   PyOrientConnectionException,
                                   PyOrientSecurityAccessException)

from ebel.config import get_config_as_dict, write_to_config
from ebel.constants import TerminalFormatting as TF
from ebel.defaults import CONN_STR_DEFAULT

logger = logging.getLogger(__name__)


def orientdb_connection_works(server: str, port: int, name: str, user: str, password: str) -> bool:
    """Check if the connection to OrientDB works."""
    try:
        client = OrientDB(server, port)
        client.set_session_token(True)
        client.db_open(name, user, password)
        works = True

    except Exception as inst:
        print(type(inst))
        works = False

    return works


def get_orientdb_client(
    server: str,
    port: int,
    name: str,
    user: str,
    password: str,
    root_password: Optional[str] = None,
    user_reader: Optional[str] = None,
    user_reader_password: Optional[str] = None,
) -> OrientDB:
    """Attempts to connect to the OrientDB client. This is currently done by using session tokens."""
    # TODO PyOrientStorageException occurs with newer ODB versions
    server = get_config_as_dict()["DEFAULT_ODB"]["server"]
    client = OrientDB(server, port)

    # First try connect as admin user if this fails connect root_user from config,
    # if this fails ask for root password and create user with admin in database
    try:
        client.set_session_token(True)
        client.db_open(name, user, password)

    except (
        PyOrientCommandException,
        PyOrientConnectionException,
        PyOrientSecurityAccessException,
    ):
        client.set_session_token(True)
        root_passwd_correct = False
        while not root_passwd_correct:
            odb_root_question = f"{TF.QUESTION}OrientDB root password (to create database and users): {TF.RESET}"
            root_password = root_password or getpass.getpass(odb_root_question)
            try:
                client.connect("root", root_password)
                root_passwd_correct = True
            except (PyOrientConnectionException, PyOrientSecurityAccessException):
                logger.error(f"Connection problem to OrientDB server {server}:{port}")
                print(
                    f"Please make sure the OrientDB server is running, port ({port}), "
                    f"as well server ({server}) and root password are correct"
                )
                sys.exit()

        if not client.db_exists(name):
            client.db_create(name)
            logger.info(f"Create database '{name}'")
            client.db_open(name, "root", root_password)
            # create user with admin rights
            client.command(f"CREATE USER {user} IDENTIFIED BY {password} ROLE admin")
            # create a reader
            if user_reader and user_reader_password:
                client.command(f"CREATE USER {user_reader} IDENTIFIED BY {user_reader_password} ROLE reader")
            client.close()
            # reopen with new user and password
            client = OrientDB(server, int(port))
            client.set_session_token(True)
            client.db_open(name, user, password)
        else:
            client.db_open(name, "root", root_password)
            # admin
            admin_user_exists_sql = (
                "Select true as admin_user_exists from OUser "
                f"where name = '{user}' and status='ACTIVE' and 'admin' in roles.name"
            )
            admin_user_exists = client.command(admin_user_exists_sql)
            if admin_user_exists:
                print("Update password for OrientDB admin")
                client.command(f"UPDATE OUser SET password = '{password}' WHERE name = '{user}'")
            else:
                print("Create password for OrientDB reader")
                client.command(f"CREATE USER {user} IDENTIFIED BY {password} ROLE admin")

            # reader
            if user_reader and user_reader_password:
                reader_user_exists_sql = (
                    "Select true as admin_user_exists from OUser "
                    f"where name = '{user_reader}' and status='ACTIVE' and 'reader' in roles.name"
                )
                reader_user_exists = client.command(reader_user_exists_sql)
                if reader_user_exists:
                    client.command(
                        f"UPDATE OUser SET password = '{user_reader_password}' WHERE name = '{user_reader}'"
                    )
                else:
                    client.command(f"CREATE USER {user_reader} IDENTIFIED BY {user_reader_password} ROLE admin")

    return client


def set_mysql_interactive() -> tuple:
    """Interactive mode to setup MySQL database."""
    print(
        "Interactive mode\n \
          ================\n \
          1st setup of db and user with root:\n"
    )
    host = input("host[localhost]:") or "localhost"
    user = input("ebel user[ebel]:") or "ebel"
    password = getpass.getpass(prompt="ebel password[ebel]:") or "ebel"
    db = input("database name[ebel]") or "ebel"
    print(
        "If you want to setup the database automatically,\n \
          then type in the root password, otherwise nothing"
    )
    root_pwd = getpass.getpass(prompt="root password (only for 1st setup):")

    if root_pwd:
        root_host = getpass.getpass(prompt="IP or name mysql server [localhost]:") or "localhost"
        conn = pymysql.connect(host=root_host, user="root", password=root_pwd)
        c = conn.cursor()
        db_exists = c.execute(f"show databases like '{db}'")

        if not db_exists:
            c.execute(f"CREATE DATABASE {db} CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci")
        else:
            logger.warning(f"Database '{db}' already exists!")

        user_exists = c.execute(f"Select 1 from mysql.user where User='{user}'")

        if not user_exists:
            sql = "CREATE USER '{}'@'{}' IDENTIFIED BY '{}'".format(
                user,
                host,
                password,
            )
            c.execute(sql)
        else:
            logger.warning(f"Database '{db}' already exists!")

        privileges_exists = c.execute(f"Select 1 from mysql.db where User='{user}' and Db='{db}'")
        if not privileges_exists:
            c.execute(f"GRANT ALL PRIVILEGES ON {db}.* TO '{user}'@'%'  IDENTIFIED BY '{password}'")
        else:
            logger.warning(f"User already has privileges for database '{db}'")

        c.execute("FLUSH PRIVILEGES")

    return host, user, password, db


def set_mysql_connection(
    host: str = "localhost",
    user: str = "ebel_user",
    password: str = "ebel_passwd",
    db: str = "ebel",
    port: Union[str, int] = "3306",
    charset: str = "utf8mb4",
):
    """Set the connection using MySQL Parameters.

    Parameters
    ----------
    host : str
        MySQL database host.
    user : str
        MySQL database user.
    password : str
        MySQL database password.
    db : str
        MySQL database name.
    port : Union[str,int]
        The port of the MySQL database.
    charset : str
        MySQL database charset.

    Returns
    -------
    str
        SQLAlchemy MySQL connection string.

    """
    connection_string = f"mysql+pymysql://{user}:{password}@{host}/{db}:{port}?charset={charset}"
    set_connection(connection_string)

    return connection_string


def set_always_create_new_db(always_create_new_db: bool = True) -> None:
    """Set in configuration option `always_create_new_db`.

    Parameters
    ----------
    always_create_new_db : bool
        Option `always_create_new_db` in section `database` in config file.

    """
    write_to_config("database", "always_create_new", str(always_create_new_db))


def set_connection(connection: str = CONN_STR_DEFAULT) -> None:
    """Set the connection string for SQLAlchemy.

    Parameters
    ----------
    connection: str
        SQLAlchemy connection string.

    """
    write_to_config("DATABASE", "sqlalchemy_connection_string", connection)



"""Protein Atlas module."""
from typing import Dict, Optional

import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.manager.orientdb import odb_meta, urls
from ebel.manager.orientdb.constants import PROTEIN_ATLAS
from ebel.manager.rdbms.models import protein_atlas
from ebel.tools import get_file_path


class ProteinAtlas(odb_meta.Graph):
    """Protein Atlas (https://www.proteinatlas.org/about/download)."""

    def __init__(self, client: OrientDB = None):
        """Init ProteinAtlas."""
        self.client = client
        self.biodb_name = PROTEIN_ATLAS
        self.urls = {
            protein_atlas.ProteinAtlasNormalTissue.__tablename__: urls.PROTEIN_ATLAS_NORMAL_TISSUE,
            protein_atlas.ProteinAtlasSubcellularLocation.__tablename__: urls.PROTEIN_ATLAS_SUBCELLULAR_LOCATIO,
            protein_atlas.ProteinAtlasRnaTissueConsensus.__tablename__: urls.PROTEIN_ATLAS_RNA_CONSENSUS,
            protein_atlas.ProteinAtlasRnaBrainGtex.__tablename__: urls.PROTEIN_ATLAS_RNA_GTEX_BRAIN,
            protein_atlas.ProteinAtlasRnaBrainFantom.__tablename__: urls.PROTEIN_ATLAS_RNA_FANTOM_BRAIN,
            protein_atlas.ProteinAtlasRnaMouseBrainAllen.__tablename__: urls.PROTEIN_ATLAS_RNA_MOUSE_BRAIN_ALLEN,
        }
        super().__init__(urls=self.urls, tables_base=protein_atlas.Base, biodb_name=self.biodb_name)

    def __len__(self):
        return self.session.query(protein_atlas.ProteinAtlasNormalTissue).count()

    def __contains__(self, ensembl_id) -> bool:
        count = (
            self.session.query(protein_atlas.ProteinAtlasNormalTissue)
            .filter(protein_atlas.ProteinAtlasNormalTissue.gene == ensembl_id)
            .count()
        )
        return bool(count)

    def insert_data(self) -> Dict[str, int]:
        """Insert NCBI gene data.

        :return Dict[str, int]: Inserted metadata.
        """
        self.recreate_tables()
        inserts = dict()
        inserts[protein_atlas.ProteinAtlasNormalTissue.__tablename__] = self._insert_normal_tissue()
        inserts[protein_atlas.ProteinAtlasSubcellularLocation.__tablename__] = self._insert_subcellular_location()
        inserts[protein_atlas.ProteinAtlasRnaTissueConsensus.__tablename__] = self._insert_rna_tissue_consensus()
        inserts[protein_atlas.ProteinAtlasRnaBrainGtex.__tablename__] = self._insert_rna_brain_gtex()
        inserts[protein_atlas.ProteinAtlasRnaBrainFantom.__tablename__] = self._insert_rna_brain_fantom()
        inserts[protein_atlas.ProteinAtlasRnaMouseBrainAllen.__tablename__] = self._insert_rna_mouse_brain_allen()
        return inserts

    def __insert_table(self, model, use_cols=None, sep: str = "\t", chunksize: Optional[int] = None) -> int:
        """Generic method to insert data.

        :param model: NCBI SQLAlchemy model
        :param use_cols: Column names to be used in file
        :param sep: Column separator
        :return int: Number of inserts.
        """
        table = model.__tablename__
        file_path = get_file_path(self.urls[table], self.biodb_name)
        dfs = pd.read_csv(file_path, sep=sep, usecols=use_cols, chunksize=chunksize)
        if chunksize is None:
            dfs = [dfs]

        number_of_inserts = 0

        for df in dfs:
            self._standardize_dataframe(df)
            df.index += 1
            df.index.rename("id", inplace=True)
            df.to_sql(table, self.engine, if_exists="append")
            number_of_inserts += df.shape[0]

        return number_of_inserts

    def _insert_normal_tissue(self) -> int:
        """Insert normal tissue.

        :return int: Number of inserts
        """
        return self.__insert_table(protein_atlas.ProteinAtlasNormalTissue)

    def _insert_subcellular_location(self) -> int:
        """Insert subcellular location.

        :return int: Number of inserts
        """
        return self.__insert_table(protein_atlas.ProteinAtlasSubcellularLocation)

    def _insert_rna_tissue_consensus(self) -> int:
        """Insert RNA tissue consensus.

        :return int: Number of inserts
        """
        return self.__insert_table(protein_atlas.ProteinAtlasRnaTissueConsensus)

    def _insert_rna_brain_gtex(self) -> int:
        """Insert RNA brain GTEx.

        :return int: Number of inserts
        """
        return self.__insert_table(protein_atlas.ProteinAtlasRnaBrainGtex)

    def _insert_rna_brain_fantom(self) -> int:
        """Insert RNA brain Fantom.

        :return int: Number of inserts
        """
        return self.__insert_table(protein_atlas.ProteinAtlasRnaBrainFantom)

    def _insert_rna_mouse_brain_allen(self) -> int:
        """Insert RNA mouse brain Allen.

        :return int: Number of inserts
        """
        return self.__insert_table(protein_atlas.ProteinAtlasRnaMouseBrainAllen)

    def get_tissues_by_ensembl_id(self, ensembl_gene_id):
        """Return tissues for a given ensembl ID."""
        level_exixs = protein_atlas.ProteinAtlasNormalTissue.level.in_(["Medium", "High", "Low"])
        columns = (
            protein_atlas.ProteinAtlasNormalTissue.tissue,
            protein_atlas.ProteinAtlasNormalTissue.cell_type,
            protein_atlas.ProteinAtlasNormalTissue.level,
        )
        data = [
            x
            for x in self.session.query(*columns)
            .filter(level_exixs)
            .filter_by(gene=ensembl_gene_id, reliability="Approved")
        ]
        tissues = {}
        for d in data:
            if d.tissue not in tissues:
                tissues[d.tissue] = {d.level: [d.cell_type]}
            else:
                if d.level not in tissues[d.tissue]:
                    tissues[d.tissue][d.level] = [d.cell_type]
                else:
                    tissues[d.tissue][d.level].append(d.cell_type)
        return tissues

    def update_interactions(self) -> int:
        """Update edges with Protein Atlas metadata (human)."""
        # TODO: implement also for other species
        match = """Select
                @rid.asString() as rid,
                hgnc.@rid.asString() as hgnc_id,
                hgnc.ensembl_gene_id as ensembl_gene_id,
                namespace,
                name,
                uniprot,
                label
            from protein
            where
                pure=true and
                namespace='HGNC' and
                both('bel_relation').size()>=1 and
                hgnc.ensembl_gene_id IS NOT NULL"""

        rid_ensembl_gene_ids = {x.oRecordData["ensembl_gene_id"]: x for x in self.execute(match)}

        self.execute("Delete EDGE has_located_protein where levels IS NOT NULL")

        location_rid_cache = {x["bel"]: x["rid"] for x in self.query_class("location", columns=["bel"])}

        for ensembl_gene_id, data in tqdm(rid_ensembl_gene_ids.items()):
            ns_location = "PROTEIN_ATLAS"
            pure_protein = data.oRecordData
            ns = pure_protein["namespace"]
            name = pure_protein["name"]

            value_dict = {
                "namespace": ns,
                "name": name,
                "hgnc": pure_protein["hgnc_id"],
                "involved_genes": [pure_protein["name"]],
                "involved_other": [],
                "species": 9606,
                "uniprot": pure_protein.get("uniprot"),
                "label": pure_protein.get("label"),
            }

            tissues = self.get_tissues_by_ensembl_id(ensembl_gene_id)

            for tissue, levels in tissues.items():
                value_dict_located = value_dict.copy()
                location_bel = f'loc({ns_location}:"{tissue}")'
                bel = f'p({ns}:"{name}",{location_bel})'
                value_dict_located.update(bel=bel)
                protein_located_rid = self.get_create_rid(
                    class_name="protein", value_dict=value_dict_located, check_for="bel"
                )

                self.create_edge(
                    class_name="has_located_protein",
                    from_rid=pure_protein["rid"],
                    to_rid=protein_located_rid,
                    value_dict={"levels": levels},
                )

                if location_bel in location_rid_cache:
                    location_rid = location_rid_cache[location_bel]
                else:
                    location_rid = self.get_create_rid(
                        class_name="location",
                        value_dict={
                            "namespace": ns_location,
                            "name": tissue,
                            "bel": location_bel,
                        },
                        check_for="bel",
                    )
                    location_rid_cache[location_bel] = location_rid

                self.create_edge("has__location", from_rid=protein_located_rid, to_rid=location_rid)



"""KEGG RDBMS model definition."""
from sqlalchemy import Column, Integer, String, Text
from sqlalchemy.ext.declarative import declarative_base

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class Mirtarbase(Base):
    """Class definition for the mirtarbase table."""

    __tablename__ = "mirtarbase"
    id = Column(Integer, primary_key=True)

    mi_rtar_base_id = Column(String(20))
    mi_rna = Column(String(50))
    species_mi_rna = Column(String(50), index=True)
    target_gene = Column(String(50), index=True)
    target_gene_entrez_id = Column(Integer)
    species_target_gene = Column(String(50), index=True)
    experiments = Column(Text)
    support_type = Column(String(50), index=True)
    references_pmid = Column(Integer)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)
